import json
from logging import LogRecord
import os
import subprocess
from typing import List, Tuple, Union, Dict, Any
from uuid import UUID, uuid4

from v0.relationalai.debugging import Span, encode_log_message, filter_span_attrs, logger
from v0.relationalai.util.constants import SPAN_ATTR_ALLOW_LIST, SPAN_TYPES_KEYS
from v0.relationalai.util.format import default_serialize
from v0.relationalai.clients.config import Config

TRACE_ID = uuid4()

class SpanTracker:
    """
    Abstract base class for tracking spans produced as log messages. 
    Provides methods to handle the start and end of spans, format span messages, and manage span attributes.
    """

    def __init__(self, trace_id: UUID, span_types_to_skip=None, log_span_attrs_as_str=True):
        self.trace_id = trace_id
        self._set_span_types_to_skip(span_types_to_skip)
        self._log_span_attrs_as_str = log_span_attrs_as_str
        self.open_spans = {}
        self.skipped_spans = set()

    def handle_span_start(self, msg: dict) -> Union[None,Span]:
        span: Span = msg["span"]
        assert_key_exists_for_span_type(span)
        if span.type in self.span_types_to_skip:
            self.skipped_spans.add(str(span.id))
            return None
        self.open_spans[str(span.id)] = span
        return span

    def handle_span_end(self, msg: dict) -> Union[Tuple[dict, str, dict], None]:
        if str(msg["id"]) in self.skipped_spans:
            return None
        span = self.open_spans.pop(msg["id"])
        joined_key_path = get_span_path_as_str(span)
        end_attrs = get_span_allowed_attributes_values(span, msg["end_attrs"])
        combined_attributes = {
            **span.attrs, # span_start attrs
            **end_attrs, # span_end attrs
            "key_path": joined_key_path,
            "elapsed_s": (span.end_timestamp - span.start_timestamp).total_seconds(),
        }
        filtered_attrs = filter_span_attrs(combined_attributes)
        formatted_span_msg = format_span_msg(self.trace_id, span, filtered_attrs, attrs_as_str=self._log_span_attrs_as_str)
        return formatted_span_msg, joined_key_path, filtered_attrs

    def _set_span_types_to_skip(self, span_types_to_skip):
        if span_types_to_skip is None:
            self.span_types_to_skip = set()
        elif isinstance(span_types_to_skip, str):
            self.span_types_to_skip = {span_types_to_skip}
        elif isinstance(span_types_to_skip, (list, tuple)):
            self.span_types_to_skip = set(span_types_to_skip)
        elif isinstance(span_types_to_skip, set):
            self.span_types_to_skip = span_types_to_skip
        else:
            raise ValueError(f"Invalid span_types_to_skip: {span_types_to_skip}, type: {type(span_types_to_skip)}")

def format_span_json(trace_id: UUID, span: Span, attrs: dict, attrs_as_str:bool=True) -> dict:
    span_key = get_span_value_from_key(span)
    if attrs_as_str:
        a = json.dumps(attrs, default=default_serialize)
    else:
        a = encode_log_message(attrs)

    span_json = span.to_json()
    extra_attrs = {
        "trace_id": trace_id,
        "key": str(span_key),
        "attrs": a
    }
    formatted_span_msg = {**span_json, **extra_attrs}
    return formatted_span_msg

#region Span formatting and filtering functions
def format_span_msg(trace_id: UUID, span: Span, attrs: dict, attrs_as_str:bool=True) -> dict:
    span_key = get_span_value_from_key(span)
    if attrs_as_str:
        a = json.dumps(attrs, default=default_serialize)
    else:
        a = encode_log_message(attrs)

    span_json = span.to_json()
    extra_attrs = {
        "trace_id": str(trace_id) if attrs_as_str else trace_id,
        "key": str(span_key) if attrs_as_str else span_key,
        "attrs": a
    }
    formatted_span_msg = {**span_json, **extra_attrs}
    return formatted_span_msg

def get_span_allowed_attributes_values(span: Span, attrs=None) -> dict:
    # For an input span with attrs (or instead for an `attrs` dict), return a dict with only the allowed attributes for the span type
    attrs = attrs or span.attrs
    out = {}
    allowed_attributes = SPAN_ATTR_ALLOW_LIST.get(span.type, set())
    for attr in allowed_attributes:
        if attr in attrs:
            out[attr] = attrs[attr]
    return out

def assert_key_exists_for_span_type(span: Span):
    if span.type in SPAN_TYPES_KEYS:
        key = SPAN_TYPES_KEYS[span.type]
        if key not in span.attrs:
            raise ValueError(f"Attribute {key} is required for spans of type {span.type}, but wasn't provided. Provided attributes:\n{span.attrs}")

def get_span_key_path(span: Span) -> str:
    # Return the key path of the current span
    value = get_span_value_from_key(span)
    if value is None:
        return span.type
    return f"{span.type}({value})"

def get_span_path_as_list(span: Span) -> List[Span]:
    # Return the path from the root to the current span
    path = []
    cur = span
    while cur is not None:
        path.append(cur)
        cur = cur.parent
    path.reverse()
    return path

def get_span_path_as_str(span: Span) -> str:
    # Return a string representation of the path from the root to the current span
    # formerly format_key_path function
    return '.'.join([get_span_key_path(span) for span in get_span_path_as_list(span)])

def get_span_value_from_key(span: Span):
    if span.type in SPAN_TYPES_KEYS:
        key = SPAN_TYPES_KEYS[span.type]
        if key in span.attrs:
            return span.attrs[key]
        raise ValueError(f"Key {key} not found in span of type {span.type}\n{span.attrs}")
    return None

def record_is_span(record: LogRecord) -> bool:
    return (
        isinstance(record.msg, dict) and
        "event" in record.msg and
        record.msg["event"] in ("span_start", "span_end")
    )

def _get_local_git_info() -> Dict[str, str]:
    try:
        branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip().decode('utf-8')
        sha = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode('utf-8')
        repo = subprocess.check_output(['git', 'config', '--get', 'remote.origin.url']).strip().decode('utf-8').rsplit('/', maxsplit=1)[-1].replace('.git', '')
        return {
            'commit': sha,
            'branch': branch,
            'repo': repo,
        }
    except subprocess.CalledProcessError as e:
        logger.warning(f"Could not get git info: {e}")
        return {
            'commit': 'local',
            'branch': 'local',
            'repo': 'local',
        }

def _get_git_info() -> Dict[str, str]:
    if os.getenv("GITHUB_REF_NAME") is not None:
        return {
            'commit': os.getenv("GITHUB_SHA", ""),
            'branch': os.getenv("GITHUB_REF_NAME", ""),
            'repo': os.getenv("GITHUB_REPOSITORY", ""),
        }
    return _get_local_git_info()

def _get_masked_config() -> Dict[str, Any]:
    """Return a config with all sensitive keys (secrets and passwords) removed, such that it can be safely logged"""
    config = Config()
    masked_props = {
        k: v for k, v in config.props.items()
        if not isinstance(v, str)
        or not any(prop_key_pattern in k for prop_key_pattern in ["secret", "password"])
    }
    return masked_props

def _get_cloud_provider() -> Dict[str, str]:
    provider = os.getenv("RAI_CLOUD_PROVIDER")
    if not provider:
        config = Config()
        provider = config.get("platform")
    assert provider, "Could not retrieve cloud provider from environment or config"
    return {"platform": provider}

def get_root_span_attrs(get_full_config: bool = False) -> Dict[str, Any]:
    git_info = _get_git_info()
    cloud_provider = _get_cloud_provider()
    masked_config = _get_masked_config() if get_full_config else {}
    attrs = {
        **git_info,
        **cloud_provider,
        **masked_config
    }
    return attrs

#endregion



from v0.relationalai.early_access.dsl.core.exprs import contextStack


class Namespace:
    top: 'Namespace'
    std: 'Namespace'

    def __enter__(self):
        contextStack.push(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()
        pass

    def __init__(self, nm, par=None):
        self._name = nm
        self._parent = par

    def name(self):
        return self._name

    def qualified_name(self):
        if self._parent is None or self._parent == Namespace.top:
            return self.name()
        else:
            # TODO: [VAMI] We changed `::` to `__` to spead up E2E testing. We should change it back to `::` before release.
            return f"{self._parent.name()}__{self.name()}"


Namespace.top = Namespace("")
Namespace.std = Namespace("std", Namespace.top)



from typing import List, Dict, Set, Tuple
from v0.relationalai.debugging import logger, warn

from v0.relationalai.experimental.pathfinder.filter import EdgeLabel
from v0.relationalai.experimental.pathfinder.transition import (
    EdgeTransition, NodeTransition, simulate
)

# =========================================================================================
# Finite Automaton
# =========================================================================================
#
# An NFA consists of:
# - states -- a finite set of states (integers)
# - init ⊆ states -- a set of initial states
# - delta: states → states → {EdgeTransition} -- transition function
# - final: states → {NodeTransition} -- final states function
#
# We introduce the NFA because a number of simplification manipulations are easier to
# perform on an NFA than on a Glushkov outline. The NFA is obtained from the Glushkov
# outline simply by adding a distinguished initial state 0 and using first transitions and
# last transitions to extend the transition function and final states. The simplification
# operations only remove unnecessary states and transitions; and their main purpose is to
# remove simple cases of non-determinism due to the use of the repetition `[n:m]` in RPQs.
# We do not perform more sophisticated operations such as determinization as those may lead
# to introduction of new states and loss of correspondence between states and the positions
# in the original RPQ.
#
# ------------------------------------------------------------------------------------------
# Example
# ------------------------------------------------------------------------------------------
# Recall the RPQ `path(OddNode, star('-[A]->', EvenNode) | path('<-[B]-', Node)` and its
# Glushkov automaton outline:
#
#  * epsilon:
#    ⋅ ⊨⸨OddNode⸩⇒
#  * positions (label):
#      1 -⟨A⟩→
#      2 ←⟨B⟩-
#  * first:
#      → 1:
#        ⋅ ⊨⸨OddNode⸩=⟦-⟨A⟩→⟧=⸨⸩⇒
#      → 2:
#        ⋅ ⊨⸨⸩=⟦←⟨B⟩-⟧=⸨⸩⇒
#  * follows:
#     - 1
#       → 1:
#         ⋅ ⊨⸨EvenNode⸩=⟦-⟨A⟩→⟧=⸨⸩⇒
#  * last:
#     - 1:
#       ⋅ ⊨⸨EvenNode⸩⇒
#     - 2:
#       ⋅ ⊨⸨Node⸩⇒
#
# The corresponding (reduced) finite automaton follows:
#
# * states: 0 1 2
# * init: 0
# * delta:
#    - 0
#       → 1
#         ⋅ ⊨⸨OddNode⸩=⟦-⟨A⟩→⟧=⸨⸩⇒
#       → 2
#         ⋅ ⊨⸨⸩=⟦←⟨B⟩-⟧=⸨⸩⇒
#    - 1
#       → 1
#         ⋅ ⊨⸨EvenNode⸩=⟦-⟨A⟩→⟧=⸨⸩⇒
# * final:
#    - 0
#       ⋅ ⊨⸨OddNode⸩⇒
#    - 1
#       ⋅ ⊨⸨EvenNode⸩⇒
#    - 2
#       ⋅ ⊨⸨Node⸩⇒
# =========================================================================================

class Automaton:
    def __init__(self,
                 states: Set[int],
                 init:Set[int],
                 delta: Dict[int, Dict[int, Set[EdgeTransition]]],
                 final: Dict[int, Set[NodeTransition]]):
        self.states = states
        self.init = init
        self.delta = delta
        self.final = final

    def __repr__(self) -> str:
        res = "Automaton:\n"
        res += "* states: " + ' '.join([str(s) for s in sorted(self.states)]) + "\n"
        res += "* init: " + ' '.join([str(s) for s in sorted(self.init)]) + "\n"
        res += "* delta:\n"
        for q in sorted(self.delta):
            res += f"   - {q}\n"
            for p in sorted(self.delta[q]):
                res += f"      → {p}\n"
                for t in self.delta[q][p]:
                    res += f"        ⋅ {t}\n"
        res += "* final: \n"
        for q in sorted(self.final):
            res += f"   - {q}\n"
            for t in self.final[q]:
                res += f"      ⋅ {t}\n"
        return res[:-1]

    def __len__(self) -> int:
        return len(self.states)

    def to_dict(self) -> dict:
        return {
            "type": "Automaton",
            "states": list(sorted(self.states)),
            "init": list(sorted(self.init)),
            "delta": {
                q: {
                    p: [t.to_dict() for t in self.delta[q][p]]
                    for p in self.delta[q]
                }
                for q in self.delta
            },
            "final": {
                q: [t.to_dict() for t in self.final[q]]
                for q in self.final
            }
        }

    @property
    def alphabet(self) -> Set[EdgeLabel]:
        res = set()
        for q in self.delta:
            for p in self.delta[q]:
                for t in self.delta[q][p]:
                    res.add(t.get_label())
        return res

    # Methods to query the automaton regardless of whether it is complete or not
    def get_edge_transitions(self, q: int, p: int) -> Set[EdgeTransition]:
        return self.delta.get(q, dict()).get(p, set())

    def get_edge_labels(self, q: int, p: int) -> Set[EdgeLabel]:
        return {t.get_label() for t in self.get_edge_transitions(q, p)}

    def has_edge_transitions(self, q: int, p: int) -> bool:
        return len(self.get_edge_transitions(q, p)) > 0

    def get_outbound_states(self, q: int) -> Set[int]:
        return {p for p in self.states if self.has_edge_transitions(q, p)}

    def get_inbound_states(self, p: int) -> Set[int]:
        return {q for q in self.states if self.has_edge_transitions(q, p)}

    def get_initial_state(self) -> int:
        assert len(self.init) == 1, "Automaton is non-deterministic: multiple initial states"
        return next(iter(self.init))

    def get_final_states(self) -> Set[int]:
        return {q for q in self.states if self.is_final(q)}

    def is_final(self, q: int) -> bool:
        return len(self.get_final_transitions(q)) > 0

    def is_exit(self, q: int) -> bool:
        return self.is_final(q) and len(self.get_outbound_states(q)) == 0

    def get_final_transitions(self, q: int) -> Set[NodeTransition]:
        return self.final.get(q, set())

    # Returns a list of strings explaining causes of non-determinism
    def determinism_report(self) -> List[str]:
        report = []

        # this won't be happening if we work with Glushkov automata from RPQ
        if len(self.init) > 1:
            report.append(f"Multiple initial states: {', '.join(map(str, self.init))}")

        def complain(q: int, lab: EdgeLabel, T: Set[Tuple[EdgeTransition, int]]) -> str:
            if q == 0:
                complaint = f"Ambiguous initial transitions for label {lab}:\n"
            else:
                complaint = f"Ambiguous transitions from {q} with label {lab}:\n"
            for (t, p) in T:
                complaint += f"  -> to position {p} (with transition {t})\n"
            return complaint

        for q in self.states:
            q_labels: Dict[EdgeLabel, Set[Tuple[EdgeTransition, int]]] = dict()
            for p in self.get_outbound_states(q):
                for t in self.delta[q][p]:
                    lab = t.get_label()
                    if lab not in q_labels:
                        q_labels[lab] = set()
                    q_labels[lab].add((t, p))

            for lab in q_labels:
                if len(q_labels[lab]) > 1:
                    report.append(complain(q, lab, q_labels[lab]))

        return report

    def is_deterministic(self) -> bool:
        return len(self.determinism_report()) == 0

    # Checks if the automaton has no reachable final states i.e. empty language.
    def is_void(self) -> bool:
        reachable: Set[int] = set()
        frontier: Set[int] = self.init.copy()
        while len(frontier) > 0:
            q = frontier.pop()
            if q in reachable:
                continue
            reachable.add(q)
            for p in self.get_outbound_states(q):
                if p in reachable | frontier:
                    continue
                frontier.add(p)

        return len(reachable & self.get_final_states()) == 0

    # Checks whether the automaton is grounded, i.e., it accepts only a finite number of
    # paths in given a finite graph. An automaton can accept an infinite number of
    # single-node paths if the initial state is accepting with an final ε-transition that
    # does not check for node label.
    def is_grounded(self) -> bool:
        return all(t.is_grounded() for q in self.init for t in self.get_final_transitions(q))

    # ----------------------------------------------------------------------------
    # Reduction
    # ----------------------------------------------------------------------------

    # Simplifies the automaton by removing unnecessary states and transitions.
    def reduce(self, prune_states_only=False):
        while True:
            pruned = False

            logger.debug("Pruning iteration for")
            logger.debug(self)

            pruned |= self.prune_unreachable_states()
            pruned |= self.prune_unproductive_states()

            if not prune_states_only:
                pruned |= self.prune_redundant_transitions()

            if not pruned:
                break

        logger.debug("Final automaton after reduction:")
        logger.debug(self)

    # Prunes unreachable states: returns True if any states were pruned.
    def prune_unreachable_states(self) -> bool:
        reachable: Set[int] = set()
        frontier: Set[int] = self.init.copy()
        while len(frontier) > 0:
            q = frontier.pop()
            if q in reachable:
                continue
            reachable.add(q)
            for p in self.get_outbound_states(q):
                if p in reachable | frontier:
                    continue
                frontier.add(p)

        unreachable = self.states - reachable

        logger.debug('Unreachable states:')
        logger.debug(list(sorted(unreachable)))

        self.prune(redundant_states=unreachable)

        return len(unreachable) > 0

    # Prunes unproductive states: returns True if any states were pruned.
    def prune_unproductive_states(self, debug: bool=True) -> bool:
        productive: Set[int] = set()
        frontier = {q for q in self.states if self.is_final(q)}
        while len(frontier) > 0:
            q = frontier.pop()
            if q in productive:
                continue
            productive.add(q)
            for p in self.get_inbound_states(q):
                if p in productive | frontier:
                    continue
                frontier.add(p)

        unproductive = self.states - productive

        logger.debug('Unproductive states:')
        logger.debug(list(sorted(unproductive)))

        self.prune(redundant_states=unproductive)

        return len(unproductive) > 0

    # Uses simulation relation on states to identify and prune redundant transitions.
    # Returns True if any transitions were pruned.
    def prune_redundant_transitions(self, debug: bool=True) -> bool:
        # The simulation relation:
        #
        # p _simulates_ q, in code `(p,q) in simulates`, iff
        # 1) for every transition t from q to q' there is a transition t' from p to p' such
        #    that t' simulates t and p' simulates q'.
        # 2) for every final transition t from q there is a final transition t' from p such
        #    that t' simulates t.
        #
        # Property: if p simulates q, then any path accepted from q is also accepted from p.
        #           The converse does not need to hold.
        #
        # The simulation relation is reflexive and transitive but not necessarily symmetric.

        # We begin with the Cartesian product of states and iteratively remove pairs
        # whenever an argument against simulation is found.
        simulates = {(p, q) for p in self.states for q in self.states}

        logger.debug("Initial simulation:")
        logger.debug(simulation_to_dict(simulates))

        # First, we remove pairs where acceptance of empty paths is not simulated
        remove_pairs = set()
        for p in self.states:
            for q in self.get_final_states():
                if not simulate(self.get_final_transitions(p), self.get_final_transitions(q)):
                    remove_pairs.add((p, q))

        simulates -= remove_pairs

        logger.debug("Pairs removed due to final state simulation:")
        logger.debug(list(sorted(remove_pairs)))
        logger.debug("Simulation pairs after final state check:")
        logger.debug(simulation_to_dict(simulates))

        # Next, we inspect the transitions between positions
        while True:
            remove_pairs = set()
            for (p, q) in simulates:
                for next_q in self.get_outbound_states(q):
                    for t_q in self.get_edge_transitions(q, next_q):
                        t_p_found = False
                        for next_p in self.get_outbound_states(p):
                            if (next_p, next_q) not in simulates:
                                continue
                            for t_p in self.get_edge_transitions(p, next_p):
                                if t_p.simulates(t_q):
                                    t_p_found = True
                        if not t_p_found:
                            remove_pairs.add((p, q))

            logger.debug("Pairs removed in transition check iteration:")
            logger.debug(list(sorted(remove_pairs)))
            logger.debug("Simulation pairs after transition check iteration:")
            logger.debug(simulation_to_dict(simulates))

            if len(remove_pairs) == 0:
                break

            simulates -= remove_pairs


        # At this point, the simulation relation is in `simulates`

        # We identify redundant transitions using the simulation relation.

        # A final node transition t1 from q is _redundant_ if there is a final node
        # transition t2 != t1 from q such that t2 simulates t1. For final transitions, we do
        # not need to worry about cycles because if t1 is redundant due to t2 and t2 is
        # redundant due to t1, then t1 == t2.

        redundant_final_transitions: Dict[int, Set[NodeTransition]] = dict()
        for q in self.get_final_states():
            for t in self.get_final_transitions(q):
                for t2 in self.get_final_transitions(q):
                    if t == t2:
                        continue
                    if t2.simulates(t):
                        logger.debug(f"Final transition {t} from {q} is redundant because simulated by {t2}")
                        if q not in redundant_final_transitions:
                            redundant_final_transitions[q] = set()
                        redundant_final_transitions[q].add(t)

        logger.debug("Redundant final transitions:")
        logger.debug(final_transitions_to_dict(redundant_final_transitions))

        # Detecting redundant edge transitions is more subtle because cycles are possible.

        # An edge-transition t1 from p to q1 _subsumes_ (is more general than) a
        # edge-transition t2 from p to q2 if t1 simulates t2, q1 simulates q2, and q1 != q2.
        # In code `(t1, q1) in subsumes[p][(t2, q2)]`.

        subsumes: Dict[int, Dict[Tuple[EdgeTransition, int], Set[Tuple[EdgeTransition, int]]]] = {
            p: dict() for p in self.states
        }

        for q in self.states:
            for q1 in self.get_outbound_states(q):
                for t1 in self.get_edge_transitions(q, q1):
                    for q2 in self.get_outbound_states(q):
                        for t2 in self.get_edge_transitions(q, q2):
                            if t1.simulates(t2) and (q1, q2) in simulates and q1 != q2:
                                if (t2, q2) not in subsumes[q]:
                                    subsumes[q][(t2, q2)] = set()
                                subsumes[q][(t2, q2)].add((t1, q1))
        logger.debug("Subsumption:")
        logger.debug(subsumption_to_dict(subsumes))

        # An edge-transition t2 from p to q2 is _redundant_ if
        # 1) there is an edge-transition t1 from p such that t1 subsumes t2 but t2 does not
        #    subsume t1.
        # 2) there is an edge-transition t1 from p to q1 such that t1 subsumes t2 and t2
        #    subsumes t1, but q1 > q2. (breaking subsumption cycles)
        redundant_edge_transitions: Dict[Tuple[int,int], Set[EdgeTransition]]= dict()
        for q in subsumes:
            for (t2, q2) in subsumes[q]:
                for (t1, q1) in subsumes[q][(t2, q2)]:
                    if (t2, q2) not in subsumes[q].get((t1, q1), set()) or q1 > q2:
                        if (q, q2) not in redundant_edge_transitions:
                            redundant_edge_transitions[(q, q2)] = set()
                        redundant_edge_transitions[(q, q2)].add(t2)

        logger.debug("Redundant edge transitions:")
        logger.debug(edge_transitions_to_dict(redundant_edge_transitions))

        self.prune(redundant_edge_transitions=redundant_edge_transitions,
                   redundant_final_transitions=redundant_final_transitions)

        return len(redundant_edge_transitions) > 0 or len(redundant_final_transitions) > 0

    # Trims the automaton by removing redundant states and transitions
    def prune(self,
              redundant_states: Set[int] = set(),
              redundant_edge_transitions: Dict[Tuple[int,int], Set[EdgeTransition]] = dict(),
              redundant_final_transitions: Dict[int, Set[NodeTransition]] = dict()):

        states = self.states - redundant_states
        init = self.init & states
        delta = dict()
        final = dict()

        for q in states:
            for p in states:
                T = self.get_edge_transitions(q, p)
                T = T - redundant_edge_transitions.get((q, p), set())
                if len(T) > 0:
                    if q not in delta:
                        delta[q] = dict()
                    delta[q][p] = T

        for q in states:
            T = self.get_final_transitions(q)
            T = T - redundant_final_transitions.get(q, set())
            if len(T) > 0:
                final[q] = T

        self.states = states
        self.init = init
        self.delta = delta
        self.final = final


#
# Private functions for logging only
#

#
# `logger` treats any dictionary as an event. To enable logging of arbitrary data
# structures we convert them to dictionary but need to encapsulate them in a class that
# implements `to_dict` method that ensures that the dictionary is not treated as an
# event but emitted as a json fragment
#
class DictEncapsulation:
    def __init__(self, d):
        self.d = d

    def to_dict(self):
        return self.d

    def __repr__(self):
        return str(self.d)

def simulation_to_dict(simulates: Set[Tuple[int, int]]):
    d = dict()
    for (p, q) in simulates:
        if p not in d:
            d[p] = list()
        d[p].append(q)
    for p in d:
        d[p].sort()
    return DictEncapsulation(d)

def final_transitions_to_dict(final_transitions: Dict[int, Set[NodeTransition]]):
    d = dict()
    for q in final_transitions:
        if len(final_transitions[q]) == 0:
            continue
        d[q] = [t.to_dict() for t in final_transitions[q]]
    return DictEncapsulation(d)

def subsumption_to_dict(subsumes: Dict[int, Dict[Tuple[EdgeTransition, int], Set[Tuple[EdgeTransition, int]]]]):
    d = dict() # p -> [{'transition': t, 'to': q, 'subsumes': [{'transition': t0, 'to': q0}]}]

    for p in subsumes:
        if len(subsumes[p]) == 0:
            continue
        if p not in d:
            d[p] = list()
        for (t, q) in subsumes[p]:
            for (t0, q0) in subsumes[p][(t, q)]:
                d[p].append({
                    'transition': t.to_dict(),
                    'to': q,
                    'subsumes': {
                        'transition': t0.to_dict(),
                        'to': q0
                    }
                })

    return DictEncapsulation(d)


def edge_transitions_to_dict(edge_transitions: Dict[Tuple[int,int], Set[EdgeTransition]]):
    d = dict() # p -> q -> [t]

    for (p, q) in edge_transitions:
        if p not in d:
            d[p] = dict()

        if q not in d[p]:
            d[p][q] = list()

        for t in edge_transitions[(p, q)]:
            d[p][q].append(t.to_dict())

    return DictEncapsulation(d)

def prepare_automaton(pattern, options) -> Automaton:
    from v0.relationalai.experimental.pathfinder.diagnostics import (
        PathfinderSizeWarning,
        PathfinderNonDeterminismWarning,
        PathfinderVoidPatternError,
        PathfinderUngroundedPatternError
    )

    force_transition_pruning = options.get('force_transition_pruning', False)

    logger.debug("Emitting PyRel for RPQ")
    logger.debug(pattern)

    # Pattern compilation
    g = pattern.glushkov()

    logger.debug(g)

    a = g.automaton()

    logger.debug("Before reduction")
    logger.debug(a)
    logger.debug(f"Size: {len(a)}")
    logger.debug(f"Automaton is deterministic: {a.is_deterministic()}")

    prune_states_only = len(pattern) >= 50
    if force_transition_pruning:
        prune_states_only = False
    if len(pattern) >= 50:
        warn(PathfinderSizeWarning(pattern, prune_states_only))

    # Automaton reduction
    logger.debug(f"Prune states only: {prune_states_only}")
    a.reduce(prune_states_only=prune_states_only)

    logger.debug("After reduction")
    logger.debug(a)
    logger.debug(f"Size: {len(a)}")
    logger.debug(f"Automaton is deterministic: {a.is_deterministic()}")

    if not a.is_deterministic():
        warn(PathfinderNonDeterminismWarning(pattern, a, prune_states_only))

    if a.is_void():
        raise PathfinderVoidPatternError(pattern)

    if not options['suppress_groundedness_test'] and not a.is_grounded():
        raise PathfinderUngroundedPatternError(pattern)

    return a



# For builder components.
from v0.relationalai.semantics import Integer, where, define, select, min
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.one_sided_ball_repetition import ball_with_repetition
from v0.relationalai.experimental.paths.path_algorithms.one_sided_ball_upto import ball_upto

# find a (deterministic) path from src to dst inside the given ball
# where dst is at distance radius from src


def single_path(g: Graph, ball, dst, radius):
    edge = g.Edge
    Node = g.Node

    path = g.model.Relationship(f"path {{Integer}} {{v:{Node}}}")

    d = Integer.ref()
    u, v = Node.ref(), Node.ref()

    define(path(radius, dst))  # path of length 0 starts at src

    where(
        path(d, v),
        edge(u, v),
        ball(d - 1, u)
    ).define(path(d - 1, min(v, u).per(d)))

    return path


def single_walk(g: Graph, Source, Target, radius):
    Node = g.Node

    ball = ball_with_repetition(g, Source, radius)

    # choose a destination node at distance radius:
    u = Node.ref()
    dst = where(ball(radius, u), Target(u)).select(min(u))
    walk = single_path(g, ball, dst, radius)

    return walk


def single_shortest_path(g: Graph, Source, Target):
    Node = g.Node

    cand = g.model.Relationship(f"candidate {{Integer}} {{v:{Node}}}")

    ball = ball_upto(g, Source, Target)

    # choose a destination node at distance radius:
    level = Integer.ref()
    u, tgt = Node.ref(), Node.ref()
    define(cand(level, u)).where(ball(level, u), Target(u))
    radius = select(level).where(cand(level, u))
    tgt = select(min(u)).where(cand(level, u))
    path = single_path(g, ball, tgt, radius)

    return path



from v0.relationalai.semantics import Integer, define, Concept, Relationship, std


class Graph:
    def __init__(self, model, Node:Concept, Edge:Relationship):
        """
        Initialize a Graph instance with Node and Edge concepts.
        """
        self.Node = Node
        self.Edge = Edge
        self.model = model


    @classmethod
    def from_edge_list(cls, model, edge_list):
        """
        Create a Graph instance from a Model.
        The Model should have Node and Edge concepts defined.
        """

        Node = model.Concept("Node")

        if len(edge_list[0]) == 2:
            Edge = model.Relationship(f"edge4 {{{Node}}} {{{Node}}}")

            for src, dst, in edge_list:
                u = Node.new(id=src)
                v = Node.new(id=dst)

                define(u)
                define(v)
                define(Edge(u, v))
        else:
            Edge = model.Relationship(f"edge5 {{{Node}}} {{{Node}}} {{String}}")

            for src, dst, label in edge_list:
                u = Node.new(id=src)
                v = Node.new(id=dst)

                define(u)
                define(v)
                define(Edge(u, v, label))

        return cls(model, Node, Edge)


    @classmethod
    def construct_grid(cls, model, side, tp = 'directed'):
        """
        Create a grid Graph instance from a Model. The length of the side of the grid
        is given by the parameter `side`.
        The Model should have Node and Edge concepts defined.
        """
        Node = model.Concept("Node")
        Edge = model.Relationship(f"edge6 {{{Node}}} {{{Node}}}")

        n, m = Integer.ref(), Integer.ref()
        u, v = Node.ref(), Node.ref()

        define(Node.new(row = n, col = m)).where(
            n == std.range(1, side + 1, 1),
            m == std.range(1, side + 1, 1)
        )

        define(Edge(u, v)).where(
            Node(u),
            Node(v),
            u.row == v.row,
            v.col == u.col + 1
        )

        define(Edge(u, v)).where(
            Node(u),
            Node(v),
            u.col == v.col,
            v.row == u.row + 1
        )

        if tp == 'undirected':
            define(Edge(u, v)).where(Edge(v, u))

        return cls(model, Node, Edge)


    @classmethod
    def construct_diamond(cls, model, diamonds, tp = 'directed'):
        """
        Create a diamond Graph instance from a Model. The number of diamonds in the graph
        is given by the parameter `diamonds`.
        The Model should have Node and Edge concepts defined.
        """
        Node = model.Concept("Node")
        Edge = model.Relationship(f"edge7 {{{Node}}} {{{Node}}}")

        n = Integer.ref()
        u, v = Node.ref(), Node.ref()

        define(Node.new(id = n)).where(
            n == std.range(1, 3*diamonds + 2, 1)
        )

        define(Edge(u, v)).where(
            n == std.range(1, diamonds + 1, 1),
            Node(u),
            Node(v),
            u.id == 3*(n-1) + 1,
            v.id == 3*(n-1) + 2
        )

        define(Edge(u, v)).where(
            n == std.range(1, diamonds + 1, 1),
            Node(u),
            Node(v),
            u.id == 3*(n-1) + 1,
            v.id == 3*(n-1) + 3
        )

        define(Edge(u, v)).where(
            n == std.range(1, diamonds + 1, 1),
            Node(u),
            Node(v),
            u.id == 3*(n-1) + 2,
            v.id == 3*n + 1
        )

        define(Edge(u, v)).where(
            n == std.range(1, diamonds + 1, 1),
            Node(u),
            Node(v),
            u.id == 3*(n-1) + 3,
            v.id == 3*n + 1
        )

        if tp == 'undirected':
            define(Edge(u, v)).where(Edge(v, u))

        return cls(model, Node, Edge)


    @classmethod
    def construct_rake(cls, model, length, fan_out, tp = 'directed'):
        """
        Create a grid Graph instance from a Model. The length of the side of the grid
        is given by the parameter `side`.
        The Model should have Node and Edge concepts defined.
        """
        Node = model.Concept("Node")
        Edge = model.Relationship(f"edge8 {{{Node}}} {{{Node}}}")

        start, i = Integer.ref(), Integer.ref()
        u, v = Node.ref(), Node.ref()

        define(Node.new(id = i)).where(
            i == std.range(0,  (length + 1) * (fan_out + 1), 1)
        )

        define(Edge(u, v)).where(
            i == std.range(0, length, 1),
            Node(u),
            Node(v),
            u.id == i,
            v.id == i + 1
        )

        define(Edge(u, v)).where(
            i == std.range(length + 1, (length + 1) * fan_out + 1, length + 1),
            Node(u),
            Node(v),
            u.id == length,
            v.id == i
        )

        define(Edge(u, v)).where(
            start == std.range(length + 1, (length + 1) * fan_out + 2, length + 1),
            i == std.range(start, start + length, 1),
            Node(u),
            Node(v),
            u.id == i,
            v.id == i + 1
        )

        if tp == 'undirected':
            define(Edge(u, v)).where(Edge(v, u))

        return cls(model, Node, Edge)



import warnings

from v0.relationalai.semantics.internal import (
    Model, Concept, Relationship, RelationshipReading, Expression, Fragment, Error,
    String, Integer, Int64, Int128, Float, Decimal, Bool,
    Date, DateTime,
    RawSource, Hash,
    select, where, require, define, distinct, union, data,
    rank, asc, desc,
    count, sum, min, max, avg, per,
    not_, internal as builder, annotations
)

__all__ = [
    "Model", "Concept", "Relationship", "RelationshipReading", "Expression", "Fragment", "Error",
    "String", "Integer", "Int64", "Int128", "Float", "Decimal", "Bool",
    "Date", "DateTime",
    "RawSource", "Hash",
    "select", "where", "require", "define", "distinct", "union", "data",
    "rank", "asc", "desc",
    "count", "sum", "min", "max", "avg", "per",
    "not_", "builder", "annotations"
]

warnings.warn(
    "relationalai.early_access.builder.* is deprecated. "
    "Please migrate to relationalai.semantics.*",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
import sys
import os

from .generic import GenericEnvironment

class TerminalEnvironment(GenericEnvironment):
    @classmethod
    def detect(cls):
        """Detect if we're running in a console/terminal environment."""
        # First, check if we're in CI - if so, we're not in a terminal
        from .ci import CIEnvironment
        if CIEnvironment.detect():
            return False

        # Check for actual terminal/console environment
        # Must have TTY attached and be interactive
        if (hasattr(sys, 'stdin') and sys.stdin.isatty() and
            hasattr(sys, 'stdout') and sys.stdout.isatty()):

            # Check for common terminal environment variables
            terminal_vars = ['TERM', 'COLORTERM', 'SSH_CLIENT', 'SSH_CONNECTION']
            if any(os.getenv(var) for var in terminal_vars):
                return True

            # Check if we have an interactive shell (ps1/ps2 attributes)
            if hasattr(sys, 'ps1') or hasattr(sys, 'ps2'):
                return True

        return False



import warnings

from v0.relationalai.semantics.std import range, hash, cast, datetime as dates, math, strings, decimals, integers, pragmas, \
    constraints, uuid_to_string

__all__ = [
    "range",
    "hash",
    "cast",
    "dates",
    "math",
    "strings",
    "decimals",
    "integers",
    "pragmas",
    "constraints",
    "uuid_to_string"
]

warnings.warn(
    "relationalai.early_access.builder.std.* is deprecated. "
    "Please migrate to relationalai.semantics.std.*",
    DeprecationWarning,
    stacklevel=2,
)



pragmas_to_lqp = {
    "rule_reasoner_sem_vo": "rel_primitive_force_faq_var_order",
    "rule_reasoner_phys_vo": "rel_primitive_force_var_order",
}

def pragma_to_lqp_name(name: str) -> str:
    if name in pragmas_to_lqp:
        return pragmas_to_lqp[name]
    else:
        # If we don't have a mapping for the built-in, we just pass it through as-is.
        return name



from v0.relationalai.experimental.pathfinder.diagnostics import PathfinderParameterWarning
from v0.relationalai.debugging import warn, logger
from typing import Callable
from v0.relationalai.dsl import Instance, Type, Graph


# ----------------------------------------------------------------------------------------
# Options controlling the execution of the Pathfinder PyRel integration
# ----------------------------------------------------------------------------------------
#
# The options used to control the execution of the Pathfinder PyRel integration are stored
# in a simple dictionary mapping strings to values (typically string, ints, and Booleans).
# For a streamline use, this dictionary will be a total mapping from the set of all
# possible options, with `None` used if the given option is not applicable (e.g., the number
# of paths to compute is only specified when the LIMIT selector is to be used).
#
# The options are:
# ## Pathfinder Rel config values
# * `graph_type` ∈ {'labeled', 'unlabeled'} (currently, only 'labeled' is supported).
# * `semantics` ∈ {'shortest_paths', 'walks'}.
# * `max_path_length`: Optional[int]; the maximum length of the paths when `semantics` is
#   'walks'; otherwise None.
# * `group` ∈ {'for_any' , 'for_each_source'}
# * `selector` ∈ {'all', 'single', 'limit'}
# * `path_count`: Optional[int]; the number of paths to return when `selector` is 'limit';
#   otherwise None
# * `search_strategy` ∈ {'from_source', 'from_target', 'from_both_sides'} graph exploration
#   strategy (default is 'from_both_sides');
# * `search_radius` ∈ {'unbounded', 'bounded'} toggle indicating whether graph exploration
#   should span the whole graph or stop whenever a target node is found (default is
#   'bounded');
# ## Optional input arguments
# * `source`: Optional[Type] or Optional[Callable]; optional specification of the source
#   node relation; by default (projection of) the connectivity relation is used to identify
#   potential sources.
# * `target`: Optional[Type] or Optional[Callable]; optional specification of the target
#   node relation; by default (projection of) the connectivity relation is used to identify
#   potential targets.
# * 'pathfinder_lib_version': str; the version of the Pathfinder library to use.
#    (default "0.7.0")
## Integration control options
# * `force_transition_pruning`: bool; if True, the automaton transition pruning is forced;
#   otherwise, pruning happens only if the size is not too big (<50).
# * `suppress_groundedness_test`: bool; if True, we do not raise an error when the input
#   automaton is ungrounded; default is False.
# * `model`: Optional[Graph]; the model to use for emitting the rules. If None, the
#   a default model is retrieved from the model stack (see utils.get_model).
#

DEFAULT_OPTIONS = {
    'graph_type': 'labeled',
    'semantics': 'shortest_paths',
    'max_path_length': None,
    'group': 'for_any',
    'selector': 'all',
    'path_count': None,
    'search_strategy': 'from_both_sides',
    'search_radius': 'bounded',
    'source': None,
    'target': None,
    'pathfinder_lib_version': '0.7.0',
    'force_transition_pruning': False,
    'suppress_groundedness_test': False,
    'model' : None
}


# -----------------------------------------------------------------------------------------
# The user-facing PyRel pathfinder functions, `find_paths` and `conn`, accept the following
# keyword _params_:
# * `max_path_length`: Optional[int] (default: None); the maximum length of the paths to be
#   computed. If max_path_length os None, the shortest paths are computed only.
# * `limit`: Optional[int] (default: None); whether all paths should be returned (None) or
#   only the first k paths.
# * `from_source` and `from_target`: Optional[bool] (default: False); whether the search
#   should be done from the source nodes or target nodes respectively. Only one of
#   `from_source` and `from_target` can be set to True. If both are False, the search is
#   done from both sides.
# * `for_each_source`: Optional[bool] (default: False); whether the search should be done
#   for each source node separately. If False, the search is done for any pair of source and
#   target nodes. This option is only supported for the shortest path semantics; When walks
#   semantics is selected (by using `max_path_length`), this option is ignored.
# * `unbounded`: Optional[bool] (default: False); whether the search should be
#   unbounded or bounded.
# * `source`: Optional[Type] or Optional[Callable]; the source node relation specification.
# * `target`: Optional[Type] or Optional[Callable]; the target node relation specification.
# * `force_transition_pruning`: Optional[bool] (default: False); whether the FA transition
#   pruning should be forced if the FA is too big (>50 states).
# * `suppress_groundedness_test`: bool; if True, we will not raises an error when the input
#   automaton is ungrounded; default is False.
# * `pathfinder_lib_version`: str; the version of the Pathfinder library to use.
#   (default "0.7.0")
# ----------------------------------------------------------------------------------------

PARAM_NAMES = [
    'max_path_length',
    'limit',
    'from_source',
    'from_target',
    'for_each_source',
    'unbounded',
    'source',
    'target',
    'force_transition_pruning',
    'suppress_groundedness_test',
    'pathfinder_lib_version'
]


def normalized_options(params):

    logger.info("Normalizing parameters...")
    for k, v in params.items():
        logger.info(f"  {k}: {v}")
        if k not in PARAM_NAMES:
            warn(PathfinderParameterWarning(
                f"Unknown parameter '{k}' ignored."
            ))

    # pathfinder_lib_version (grab)
    pathfinder_lib_version = params.get('pathfinder_lib_version',
                                        DEFAULT_OPTIONS['pathfinder_lib_version'])
    assert pathfinder_lib_version in ["0.7.0"], "`pathfinder_lib_version` parameter is an invalid library version"

    # graph_type
    graph_type = "labeled"

    # max_path_length (grab)
    max_path_length = None
    if 'max_path_length' in params:
        max_path_length = params['max_path_length']
        if not isinstance(max_path_length, Instance):
            assert type(max_path_length) is int and max_path_length >= 0, "`max_path_length` parameter must be a non-negative integer"


    # semantics (grab)
    if max_path_length is None:
        semantics = 'shortest_paths'
    else:
        semantics = 'walks'

    # path_count (grab)
    path_count = None
    if 'limit' in params:
        path_count = params['limit']
        if not isinstance(path_count, Instance):
            assert type(path_count) is int and path_count > 0, "`limit` parameter must be a positive integer"


    # selector (grab)
    if path_count is None:
        selector = 'all'
    elif path_count == 1:
        selector = 'single'
    else:
        selector = "limit"


    # group (grab)
    group = 'any_pair'
    assert 'for_each_source' not in params or type(params['for_each_source']) is bool, "`for_each_source` parameter must be a boolean"
    if 'for_each_source' in params and params['for_each_source']:
        group = 'for_each_source'
    # group (check)
    if group == 'for_each_source' and semantics == 'walks':
        warn(PathfinderParameterWarning(
            "The option `for_each_source=True` is ignored when "
            "the option `max_path_length` is used."
        ))
        group = 'any_pair'
    if group == 'for_each_source' and selector in ['single', 'limit']:
        warn(PathfinderParameterWarning(
            "The option `for_each_source=True` is ignored when "
            "the option `limit` is used."
        ))
        group = 'any_pair'


    # search_strategy (grab)
    search_strategy = 'from_both_sides'
    assert 'from_source' not in params or type(params['from_source']) is bool, "`from_source` parameter must be a boolean"
    assert 'from_target' not in params or type(params['from_target']) is bool, "`from_target` parameter must be a boolean"
    if 'from_source' in params and params['from_source']:
        search_strategy = 'from_source'
    elif 'from_target' in params and params['from_target']:
        search_strategy = 'from_target'
    # search_strategy (check)
    if group == 'for_each_source' and search_strategy != 'from_source':
        warn(PathfinderParameterWarning(
            "The parameter `for_each_source=True` forces the parameter `from_source=True`. "
            "Parameters selecting other search strategy, like `from_target=True`, are ignored."
        ))
        search_strategy = 'from_source'
    if (group == 'any_pair' and search_strategy == 'from_source' and
        'from_target' in params and params['from_target']):
        warn(PathfinderParameterWarning(
            "The parameter `from_source=True` supersedes the parameter `from_target=True`. "
            "The parameter `from_target=True` is ignored."
        ))


    # search_radius (grab)
    search_radius = 'bounded'
    assert 'unbounded' not in params or type(params['unbounded']) is bool, "`unbounded` parameter must be a boolean"
    if 'unbounded' in params and params['unbounded']:
        search_radius = 'unbounded'
    # search_radius (check)
    if search_radius == 'unbounded' and semantics == 'walks':
        warn(PathfinderParameterWarning(
            "The parameter `unbounded=True` is ignored when "
            "the parameter `max_path_length` is used."
        ))
        search_radius = 'bounded'
    if search_radius == 'unbounded' and search_strategy == 'from_both_sides':
        warn(PathfinderParameterWarning(
            "The parameter `unbounded=True` has effect only when used with one of the parameters "
            "`from_source=True` or `from_target=True; The parameter `unbounded=True` is ignored."
        ))
        search_radius = 'bounded'


    # source and target (grab)
    source = params.get("source", None)
    target = params.get("target", None)
    # source and target (check)
    if source is not None and not isinstance(source, (Type, Callable)):
        warn(PathfinderParameterWarning(
            "Ignoring parameter `source`: must be a Type or a function"
        ))
        source, target = None, None

    if target is not None and not isinstance(target, (Type, Callable)):
        warn(PathfinderParameterWarning(
            "Ignoring parameter `source`: must be a Type or a lambda expression"
        ))
        source, target = None, None

    # force_transition_pruning (grab)
    force_transition_pruning = params.get("force_transition_pruning", False)
    assert type(force_transition_pruning) is bool, "`force_transition_pruning` parameter must be a boolean"

    # suppress_groundedness_test (grab)
    suppress_groundedness_test = params.get("suppress_groundedness_test", False)
    assert type(suppress_groundedness_test) is bool, "`suppress_groundedness_test` parameter must be a boolean"

    # model (grab)
    model = params.get("model", None)
    assert model is None or isinstance(model, Graph), "`model` parameter must be a Graph instance"

    return {
        'graph_type': graph_type,
        'semantics': semantics,
        'group': group,
        'selector': selector,
        'search_strategy': search_strategy,
        'search_radius': search_radius,
        'source': source,
        'target': target,
        'max_path_length': max_path_length,
        'path_count': path_count,
        'pathfinder_lib_version': pathfinder_lib_version,
        'force_transition_pruning': force_transition_pruning,
        'suppress_groundedness_test': suppress_groundedness_test,
        'model': model
    }



import typing

from v0.relationalai.early_access.dsl.orm.printer import Printer, _sort_dependency_graph, InterfacePrinter
from v0.relationalai.early_access.dsl.orm.models import Model
from v0.relationalai.early_access.dsl.orm.types import Concept

SPACE_INDENT = "    "
TAB_INDENT = "\t"

class ObjectOrientedPrinter(Printer):
    _indent: str

    def __init__(self, io: typing.Optional[typing.IO[str]] = None, space_indent: bool= False):
        super().__init__(io)
        object.__setattr__(self, '_indent', SPACE_INDENT if space_indent else TAB_INDENT)

    def to_string(self, model: Model, enums_enabled: bool = False, pyi_enabled: bool = False) -> None:
        self._process_constraints(model)

        self._print_nl("from typing import Any, Sequence, Union, TypeVar")
        self._print_nl("import relationalai.semantics.internal as qb")
        self._print_nl("import relationalai.semantics.internal.internal as builder")
        self._print_nl("from relationalai.early_access.dsl.orm.models import Model")
        if pyi_enabled:
            self._print_nl("from relationalai.early_access.dsl.orm.models import Concept")
        self._print_nl("from relationalai.early_access.dsl.orm.relationships import Role")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import Range")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import RingType")
        self._print_nl("from relationalai.early_access.dsl.orm.constraints import ValueComparisonType")
        self._nl()

        self._handle_concepts_declaration(model, enums_enabled, pyi_enabled)
        self._nl()
        self._print_nl("class OntologyBase:\n"
                       f'{self._indent}"""Represents an Ontology."""')
        self._nl()
        self._print_nl(f"{self._indent}def __init__(self, model: Model):\n"
                       f"{self._indent}{self._indent}self.model = model")
        self._nl()
        self._print_nl(f"{self._indent}def generate_model_rules(self):\n"
                       f"{self._indent}{self._indent}self.model.generate_model_rules()")
        self._nl()
        if not pyi_enabled:
            self._print_nl(f"{self._indent}def Concept(self, name: str, extends: list[Any] = [], identify_by:dict[str, Any]={{}}):\n"
                           f"{self._indent}{self._indent}return self.model.Concept(name, extends, identify_by)")
        if enums_enabled:
            self._print_nl(f"{self._indent}def Enum(self):\n"
                           f"{self._indent}{self._indent}return self.model.Enum()")
        self._print_nl(f"{self._indent}def Relationship(self, reading: Any, short_name:str=\"\") -> qb.Relationship:\n"
                       f"{self._indent}{self._indent}return self.model.Relationship(reading, short_name)")
        self._print_nl(f"{self._indent}def Unique(self, *roles):\n"
                       f"{self._indent}{self._indent}self.model.unique(*roles)")
        self._print_nl(f"{self._indent}def Mandatory(self, role):\n"
                       f"{self._indent}{self._indent}self.model.mandatory(role)")
        self._print_nl(f"{self._indent}T = TypeVar('T', int, float, str)")
        self._print_nl(f"{self._indent}def RoleValueConstraint(self, role: Role, values: Sequence[Union[T, Range[T]]]):\n"
                       f"{self._indent}{self._indent}self.model.role_value_constraint(role, values)")
        self._print_nl(f"{self._indent}def InclusiveSubtypeConstraint(self, *concepts: Concept):\n"
                       f"{self._indent}{self._indent}self.model.inclusive_subtype_constraint(*concepts)")
        self._print_nl(f"{self._indent}def ExclusiveSubtypeConstraint(self, *concepts: Concept):\n"
                       f"{self._indent}{self._indent}self.model.exclusive_subtype_constraint(*concepts)")
        self._print_nl(f"{self._indent}def InclusiveRoleConstraint(self, *roles):\n"
                       f"{self._indent}{self._indent}self.model.inclusive_roles(*roles)")
        self._print_nl(f"{self._indent}def ExclusiveRoleConstraint(self, *role_sequences):\n"
                       f"{self._indent}{self._indent}self.model.exclusive_roles(*role_sequences)")
        self._print_nl(f"{self._indent}def RingConstraint(self, *roles):\n"
                       f"{self._indent}{self._indent}self.model.ring(*roles)")
        self._print_nl(f"{self._indent}def ValueComparisonConstraint(self, constraint_type, *roles):\n"
                       f"{self._indent}{self._indent}self.model.value_comparison(constraint_type, *roles)")
        self._print_nl(f"{self._indent}def RoleSubsetConstraint(self, *role_sequences):\n"
                       f"{self._indent}{self._indent}self.model.role_subset(*role_sequences)")
        self._print_nl(f"{self._indent}def EqualityConstraint(self, *role_sequences):\n"
                       f"{self._indent}{self._indent}self.model.equality(*role_sequences)")
        self._print_nl(f"{self._indent}def FrequencyConstraint(self, frequency, *role_sequences):\n"
                       f"{self._indent}{self._indent}self.model.frequency(frequency, *role_sequences)")
        self._print_nl(
            f"{self._indent}def CardinalityConstraint(self, concept: Concept, values: Sequence[Union[int, Range[int]]]):\n"
            f"{self._indent}{self._indent}self.model.cardinality(concept, values)")
        self._print_nl(f"{self._indent}def RoleCardinalityConstraint(self, role: Role, values: Sequence[Union[int, Range[int]]]):\n"
                       f"{self._indent}{self._indent}self.model.role_cardinality(role, values)")
        self._print_nl(f"{self._indent}def ValueConstraint(self, concept: Concept, values: Sequence[Union[T, Range[T]]]):\n"
                       f"{self._indent}{self._indent}self.model.value_constraint(concept, values)")

        self._nl()
        self._print_nl("class ORMOntology(OntologyBase):")
        self._print_nl(f'{self._indent}"""Represents an Ontology generated from an ORM file."""')
        self._nl()
        self._print_nl(f"{self._indent}def __init__(self, model: Model):")
        self._print_nl(f"{self._indent}{self._indent}super().__init__(model)")
        self._nl()
        self._handle_concepts(model, enums_enabled, pyi_enabled)
        self._nl()
        self._handle_relationships(model)

        self._nl()
        self._handle_ref_schemes(model)

        if self._unique_constraints:
            self._nl()
            self._handle_uniqueness_constraints()

        if self._mandatory_constraints:
            self._nl()
            self._handle_mandatory_constraints()

        if self._role_value_constraints:
            self._nl()
            self._handle_role_value_constraints()

        if self._inclusive_subtype_constraints or self._exclusive_subtype_constraints:
            self._nl()
            self._handle_subtype_constraints()

        if self._inclusive_role_constraints:
            self._nl()
            self._handle_inclusive_role_constraints()

        if self._exclusive_role_constraints:
            self._nl()
            self._handle_exclusive_role_constraints()

        if self._ring_constraints:
            self._nl()
            self._handle_ring_constraints()

        if self._value_comparison_constraints:
            self._nl()
            self._handle_value_comparison_constraints()

        if self._role_subset_constraints:
            self._nl()
            self._handle_role_subset_constraints()

        if self._equality_constraints:
            self._nl()
            self._handle_equality_constraints()

        if self._frequency_constraints:
            self._nl()
            self._handle_frequency_constraints()

        if self._cardinality_constraints:
            self._nl()
            self._handle_cardinality_constraints()

        if self._role_cardinality_constraints:
            self._nl()
            self._handle_role_cardinality_constraints()

        if self._value_constraints:
            self._nl()
            self._handle_value_constraints()

    def _handle_concepts_declaration(self, model: Model, enums_enabled: bool, pyi_enabled: bool) -> None:
        concepts_map = model.concepts_map()
        sorted_concepts = _sort_dependency_graph(model.concepts())
        # print Concept classes
        if pyi_enabled:
            for name in sorted_concepts:
                c = concepts_map.get(name)
                if c is None:
                    raise ValueError(f"The concept '{name}' was not declared but used as concept domain.")
                # skip enums when they are enabled for printer
                if c._is_enum() and enums_enabled:
                    continue
                extends = [f"{e._name}Concept" for e in c._extends if not e._is_primitive()]
                self._print_nl(f"class {name}Concept({', '.join(extends) if extends else 'Concept'}):")
                self._print_nl(f'{self._indent}"""Represents {name} concept."""')
            self._nl()

    def _handle_concepts(self, model: Model, enums_enabled: bool, pyi_enabled: bool) -> None:
        concepts_map = model.concepts_map()
        enums_map = model.enums_map()
        sorted_concepts = _sort_dependency_graph(model.concepts())
        for name in sorted_concepts:
            c = concepts_map.get(name)
            if c is None:
                raise ValueError(f"The concept '{name}' was not declared but used as concept domain.")
            if enums_enabled and c._is_enum():
                self._print_nl(f"{self._indent}{self._indent}self.{name} = model.Enum('{name}', {[e.name for e in enums_map[name]]})")
            else:
                extends = c._extends
                if not enums_enabled and c._is_enum():
                    # todo: derive enum type
                    # ORM adapter produces only string enums
                    extends_elements = ["qb.String"]
                else:
                    extends_elements = [f"self.{self._get_type(ext)}" if not self._get_type(ext).startswith(
                        "qb.") and not self._get_type(ext).startswith("builder.") else self._get_type(ext) for ext in
                                        extends]
                extends_str = f", extends=[{', '.join(extends_elements)}]" if extends else ""
                concept_class = name + 'Concept' if pyi_enabled else 'model.Concept'
                params = f"({'model, ' if pyi_enabled else ''}'{name}'{extends_str})"
                self._print_nl(f"{self._indent}{self._indent}self.{name} = {concept_class}{params}")

    def _handle_relationships(self, model: Model) -> None:
        for rel in model.relationships():
            # skip autogenerated 'name' Relationship for Enums
            if rel._name == 'name' and rel._parent is not None and rel._parent._is_enum():
                continue
            # print a root Relationship
            self._print_nl(f"{self._indent}{self._indent}self.{self._get_relationship_name(rel)} = "
                           f"model.Relationship('{rel._madlib}'{self._print_if_not_empty_and_not_equal('short_name', rel._passed_short_name, rel._name)})")

            # print remaining RelationshipReadings if any
            for r in rel._readings[1:]:
                self._print_nl(f"{self._indent}{self._indent}self.{self._get_relationship_name(r)} = "
                               f"self.{self._get_relationship_name(rel)}.alt('{r._madlib}'{self._print_if_not_empty_and_not_equal('short_name', r._passed_short_name, r._name)})")

    def _handle_ref_schemes(self, model: Model) -> None:
        for concept_name, concept in model.entity_types_map().items():
            if not isinstance(concept, Concept) or not concept._reference_schemes:
                continue

            for ref_scheme in concept._reference_schemes:
                rel_names = [f"self.{self._get_relationship_name(rel)}" for rel in ref_scheme]
                self._print_nl(f"{self._indent}{self._indent}self.{concept_name}.identify_by({', '.join(rel_names)})")

    def _handle_uniqueness_constraints(self) -> None:
        for constraint in self._unique_constraints:
            if not constraint.is_preferred_identifier:
                elements = [f"self.{self._get_role_name(role)}" for role in constraint.roles()]
                self._print_nl(f"{self._indent}{self._indent}self.Unique({', '.join(elements)})")

    def _handle_mandatory_constraints(self):
        for constraint in self._mandatory_constraints:
            role = constraint.roles()[0]
            role_name = self._get_role_name(role)
            self._print_nl(f"{self._indent}{self._indent}self.Mandatory(self.{role_name})")

    def _handle_role_value_constraints(self) -> None:
        for constraint in self._role_value_constraints:
            role = constraint.roles()[0]
            elements = self._get_constraint_values(constraint)
            role_name = f"self.{self._get_role_name(role)}"
            self._print_nl(f"{self._indent}{self._indent}self.RoleValueConstraint({role_name}, [{', '.join(elements)}])")

    def _handle_inclusive_role_constraints(self):
        for constraint in self._inclusive_role_constraints:
            roles = [f"self.{self._get_role_name(r)}" for r in constraint.roles()]
            self._print_nl(f"{self._indent}{self._indent}self.InclusiveRoleConstraint({', '.join(roles)})")

    def _handle_exclusive_role_constraints(self):
        self._emit_role_sequence_constraints(self._exclusive_role_constraints, "ExclusiveRoleConstraint")

    def _handle_ring_constraints(self):
        for constraint in self._ring_constraints:
            elements = [f"self.{self._get_role_name(role)}" for role in constraint.roles()]
            cst_types = [f"{tp}" for tp in constraint.types]
            self._print_nl(f"{self._indent}{self._indent}self.RingConstraint([{', '.join(cst_types)}], {', '.join(elements)})")

    def _handle_value_comparison_constraints(self):
        for constraint in self._value_comparison_constraints:
            elements = [f"self.{self._get_role_name(role)}" for role in constraint.roles()]
            self._print_nl(f"{self._indent}{self._indent}self.ValueComparisonConstraint({constraint.type}, {', '.join(elements)})")

    def _handle_frequency_constraints(self):
        for constraint in self._frequency_constraints:
            elements = [f"self.{self._get_role_name(role)}" for role in constraint.roles()]
            self._print_nl(f"{self._indent}{self._indent}self.FrequencyConstraint(({', '.join(constraint.frequency)}), {', '.join(elements)})")

    def _handle_cardinality_constraints(self):
        for constraint in self._cardinality_constraints:
            elements = self._get_constraint_values(constraint)
            self._print_nl(
                f"{self._indent}{self._indent}self.CardinalityConstraint(self.{constraint.concept()._name}, [{', '.join(elements)}])")

    def _handle_role_cardinality_constraints(self):
        for constraint in self._role_cardinality_constraints:
            elements = self._get_constraint_values(constraint)
            self._print_nl(
                f"{self._indent}{self._indent}self.RoleCardinalityConstraint(self.{self._get_role_name(constraint.roles()[0])}, [{', '.join(elements)}])")

    def _handle_value_constraints(self):
        for constraint in self._value_constraints:
            elements = self._get_constraint_values(constraint)
            self._print_nl(f"{self._indent}{self._indent}self.ValueConstraint(self.{constraint.concept()._name}, [{', '.join(elements)}])")

    def _emit_role_sequence_constraints(self, constraints, constraint_type: str):
        for constraint in constraints:
            role_sequences = [[f"self.{self._get_role_name(r)}" for r in ro_list] for ro_list in
                              constraint.role_sequences()]
            if self._is_complex_role_sequence(role_sequences):
                self._print_nl(
                    f"{self._indent}{self._indent}self.{constraint_type}([{'], ['.join(', '.join(ro_list) for ro_list in role_sequences)}])")
            else:
                self._print_nl(
                    f"{self._indent}{self._indent}self.{constraint_type}({', '.join(', '.join(ro_list) for ro_list in role_sequences)})")

    def _emit_subtype_constraints(self, constraints, constraint_type: str):
        for constraint in constraints:
            concepts = [f"self.{self._get_type(c)}" for c in constraint.concepts().values()]
            self._print_nl(f"{self._indent}{self._indent}self.{constraint_type}({', '.join(concepts)})")


EXCLUDED_RELATIONSHIPS = ["shape"]


class ObjectOrientedInterfacePrinter(InterfacePrinter):
    _indent: str

    def __init__(self, io: typing.Optional[typing.IO[str]] = None, space_indent: bool=False):
        # Set the base class field (frozen)
        super().__init__(io)
        object.__setattr__(self, '_indent', SPACE_INDENT if space_indent else TAB_INDENT)

    def to_string(self, model: Model, enums_enabled: bool = False) -> None:
        if enums_enabled:
            self._print_nl("from enum import Enum")
        self._print_nl("from typing import Union")
        self._print_nl("import relationalai.semantics.internal as qb")
        self._print_nl("import relationalai.semantics.internal.internal as builder")
        self._print_nl("from relationalai.early_access.dsl.orm.models import Concept")
        self._nl()
        if enums_enabled:
            self._handle_enums(model)
        self._handle_concepts(model, enums_enabled)

    def _handle_concepts(self, model: Model, enums_enabled) -> None:
        concepts_map = model.concepts_map()
        sorted_concepts = _sort_dependency_graph(model.concepts())
        for name in sorted_concepts:
            c = concepts_map.get(name)
            if c is None:
                raise ValueError(f"The concept '{name}' was not declared but used as concept domain.")
            if enums_enabled and c._is_enum():
                continue
            exclude_list = list(EXCLUDED_RELATIONSHIPS)
            # when enum is declared but enums are disabled print it as a Concept without "name" Relationship
            if c._is_enum():
                exclude_list.append("name")
            extends = [f"{e._name}Concept" for e in c._extends if not e._is_primitive()]
            self._print_nl(f"class {name}Concept({', '.join(extends) if extends else 'Concept'}):")
            rel_strs = self._get_declared_relationships(c, exclude_list)
            self._print_nl(f"{self._indent}pass" if len(rel_strs) == 0 else "\n".join(rel_strs))
            self._nl()
        self._print_nl("class ORMOntology:")
        for name in sorted_concepts:
            c = concepts_map.get(name)
            if c:
                if c._is_enum() and enums_enabled:
                    self._print_nl(f"{self._indent}{c._name}: {c._name}")
                else:
                    self._print_nl(f"{self._indent}{c._name}: {c._name}Concept")



from typing import Callable
from v0.relationalai.dsl import get_graph, Graph
# -----------------------------------------------------------------------------------------
# Utility functions
# -----------------------------------------------------------------------------------------

#
# Get the string representation of a lambda function (with moderate effort).
#
def get_lambda_str(f: Callable) -> str:
    try:
        if f.__name__ == '<lambda>':
            return f"λ{','.join(f.__code__.co_varnames)}:... @{hex(id(f))}"
        else:
            return f"{f.__name__}({','.join(f.__code__.co_varnames)}):... @{hex(id(f))}"
    except Exception as _:
        return str(f)

#
# Returns the model to use for emitting the logic defining paths and calling the Pathfinder.
#
def get_model(options) -> Graph:
    if options['model'] is not None:
        return options['model']
    else:
        return get_graph()



# tracer.py — stdlib-only OTEL-shaped tracer with live span_start + 200ms write throttling
from __future__ import annotations
import os, json, time, secrets, threading, atexit
from contextlib import ContextDecorator, asynccontextmanager
from contextvars import ContextVar
from typing import Optional, Callable, Any, Dict

# ---------- utils ----------
def _hex(bits: int) -> str: return secrets.token_hex(bits // 8)
def now_ns() -> int: return time.time_ns()

# Each logical execution path (thread/task) has its own span stack
_current_stack: ContextVar[list["Span"]] = ContextVar("_current_stack", default=[])

# ---------- core ----------
class Span:
    """
    Represents a span; use Tracer.span(...) / Tracer.aspan(...) to create via context manager.
    """
    __slots__ = ("tracer","name","kind","trace_id","span_id","parent_span_id",
                 "start","end","attrs","events","status","_finished")

    def __init__(self, tracer:"Tracer", name:str, kind:str, parent:"Span|None", **attrs):
        self.tracer = tracer
        self.name = name
        self.kind = kind
        self.trace_id = parent.trace_id if parent else _hex(128)
        self.span_id = _hex(64)
        self.parent_span_id = parent.span_id if parent else None
        self.start = now_ns()
        self.end: Optional[int] = None
        self.attrs: Dict[str, Any] = dict(attrs)
        self.events: list[dict] = []
        self.status = "UNSET"
        self._finished = False

    # ---- span APIs ----
    def add_event(self, name:str, **attributes):
        self.events.append({"name": name, "ts": now_ns(), "attributes": attributes})

    def set_status(self, ok: bool = True, message: Optional[str] = None):
        self.status = "OK" if ok else "ERROR"
        if message:
            self.attrs["error.message"] = message

    def record_exception(self, exc: BaseException):
        self.add_event("exception",
                       type=exc.__class__.__name__,
                       message=str(exc),
                       module=getattr(exc.__class__, "__module__", ""))
        self.set_status(False, str(exc))

    def finish(self):
        if self._finished:
            return
        self._finished = True
        self.end = now_ns()
        self.tracer._emit_full_span(self)

    def __setitem__(self, key, value):
        self.attrs[key] = value

class Tracer(ContextDecorator):
    """
    Tracer with:
      - context-managed spans (sync & async)
      - automatic parent linkage via ContextVar stack
      - live 'span_start' records + batched/throttled writes
      - OTEL-shaped final span records
    """
    def __init__(
        self,
        out_path: str = "spans.jsonl",
        service: str = "rel-modeler",
        flush_interval_ms: int = 200,
        emit_span_start: bool = True,
    ):
        self.out_path = out_path
        self.service = service
        self.emit_span_start = emit_span_start
        os.makedirs(os.path.dirname(out_path) or ".", exist_ok=True)

        self._program_span: Span | None = None

        # Buffered/throttled writer
        self._lock = threading.Lock()
        self._buffer: list[str] = []
        self._last_flush_ns: int = 0
        self._flush_interval_ns: int = flush_interval_ms * 1_000_000
        atexit.register(self.flush)

    # ---- context manager for sync code ----
    def span(self, name:str, *, kind:str="INTERNAL", **attrs):
        tracer = self

        class SpanCtx:
            def __enter__(self):
                stack = _current_stack.get()
                parent = stack[-1] if stack else (tracer._program_span if tracer._program_span and not tracer._program_span._finished else None)
                self.span = Span(tracer, name, kind, parent, **attrs)
                _current_stack.set(stack + [self.span])
                if tracer.emit_span_start:
                    tracer._emit_span_start(self.span)
                return self.span

            def __exit__(self, exc_type, exc, tb):
                try:
                    if exc_type is not None:
                        self.span.record_exception(exc)  # marks ERROR
                    elif self.span.status == "UNSET":
                        self.span.set_status(True)
                finally:
                    self.span.finish()
                    # pop this span
                    cur = _current_stack.get()
                    if cur and cur[-1] is self.span:
                        _current_stack.set(cur[:-1])
                # do not suppress exceptions
                return False

        return SpanCtx()

    # ---- async context manager ----
    @asynccontextmanager
    async def aspan(self, name:str, *, kind:str="INTERNAL", **attrs):
        stack = _current_stack.get()
        parent = stack[-1] if stack else (self._program_span if self._program_span and not self._program_span._finished else None)
        span = Span(self, name, kind, parent, **attrs)
        _current_stack.set(stack + [span])
        if self.emit_span_start:
            self._emit_span_start(span)
        try:
            yield span
            if span.status == "UNSET":
                span.set_status(True)
        except BaseException as e:
            span.record_exception(e)
            raise
        finally:
            span.finish()
            cur = _current_stack.get()
            if cur and cur[-1] is span:
                _current_stack.set(cur[:-1])

    # ---- helpers ----
    def current_span(self) -> Optional[Span]:
        stack = _current_stack.get()
        parent = stack[-1] if stack else (self._program_span if self._program_span and not self._program_span._finished else None)
        return parent

    def add_event(self, name:str, **attrs):
        s = self.current_span()
        if s:
            s.add_event(name, **attrs)

    def set_status(self, ok:bool=True, message:str|None=None):
        s = self.current_span()
        if s:
            s.set_status(ok, message)

    def traceparent(self) -> str:
        s = self.current_span()
        if not s: return ""
        # version(00)-trace_id-span_id-flags(01)
        return f"00-{s.trace_id}-{s.span_id}-01"

    def start_program(self, name: str = "program", *, kind: str = "INTERNAL", **attrs) -> Span:
        """Open one root span for the whole process; auto-finished via atexit."""
        if self._program_span and not self._program_span._finished:
            return self._program_span  # already started
        root = Span(self, name, kind, parent=None, **attrs)
        self._program_span = root
        if getattr(self, "emit_span_start", False):  # works if you added span_start earlier
            self._emit_span_start(root)
        # close on interpreter shutdown
        def _close():
            s = self._program_span
            if s and not s._finished:
                if s.status == "UNSET":
                    s.set_status(True)
                s.finish()
        atexit.register(_close)
        return root

    # ---- emitters (buffered) ----
    def _emit_span_start(self, span: Span):
        doc = {
            "type": "span_start",
            "trace_id": span.trace_id,
            "span_id": span.span_id,
            "parent_span_id": span.parent_span_id,
            "name": span.name,
            "kind": span.kind,
            "start_time_unix_nano": span.start,
            "resource": {"service.name": self.service},
            "attributes": span.attrs,  # snapshot of initial attrs
        }
        self._write(json.dumps(doc, separators=(",", ":")) + "\n")

    def _emit_full_span(self, span: Span):
        doc = {
            "trace_id": span.trace_id,
            "span_id": span.span_id,
            "parent_span_id": span.parent_span_id,
            "name": span.name,
            "kind": span.kind,
            "start_time_unix_nano": span.start,
            "end_time_unix_nano": span.end,
            "status": span.status,
            "attributes": span.attrs,
            "events": span.events,
            "resource": {"service.name": self.service},
        }
        self._write(json.dumps(doc, separators=(",", ":")) + "\n")

    def _write(self, line: str):
        now = now_ns()
        with self._lock:
            self._buffer.append(line)
            if (now - self._last_flush_ns) >= self._flush_interval_ns:
                self._flush_locked(now)

    def flush(self):
        with self._lock:
            self._flush_locked(now_ns())

    def _flush_locked(self, now: int):
        if not self._buffer:
            self._last_flush_ns = now
            return
        # Single append write for all buffered lines
        with open(self.out_path, "a", encoding="utf-8") as f:
            f.writelines(self._buffer)
        self._buffer.clear()
        self._last_flush_ns = now

# ---------- decorator ----------
def traced(name: str | None = None, *, kind: str = "INTERNAL"):
    def _wrap(fn: Callable):
        nm = name or fn.__name__
        def inner(*a, **kw):
            tracer = get_tracer()
            with tracer.span(nm, kind=kind):
                return fn(*a, **kw)
        return inner
    return _wrap

# ---------- global default (optional) ----------
_DEFAULT_TRACER: Tracer | None = None
def set_tracer(t: Tracer):  # call once at app start
    global _DEFAULT_TRACER
    _DEFAULT_TRACER = t
def get_tracer() -> Tracer:
    if _DEFAULT_TRACER is None:
        set_tracer(Tracer())  # lazy init to spans.jsonl
    assert _DEFAULT_TRACER is not None
    return _DEFAULT_TRACER

def span(name:str, *, kind:str="INTERNAL", **attrs):
    tracer = get_tracer()
    return tracer.span(name, kind=kind, **attrs)



from __future__ import annotations
from collections.abc import Iterable, Iterator, MutableSet
from typing import Callable, Generic, Hashable, TypeVar

T = TypeVar("T")

#------------------------------------------------------
# OrderedSet
#------------------------------------------------------

class OrderedSet(Generic[T], MutableSet[T]):
    __slots__ = ("_dict",)
    _dict: dict[T, None]

    def __init__(self, iterable: Iterable[T] | None = None) -> None:
        self._dict = {}
        if iterable is not None:
            self.update(iterable)

    def __contains__(self, item: object) -> bool:  # type: ignore[override]
        return item in self._dict

    def __iter__(self) -> Iterator[T]:
        return iter(self._dict.keys())

    def __len__(self) -> int:
        return len(self._dict)

    def __getitem__(self, index: int) -> T:
        if index < 0:
            index += len(self._dict)
        if index < 0 or index >= len(self._dict):
            raise IndexError("OrderedSet index out of range")
        for i, key in enumerate(self._dict):
            if i == index:
                return key
        # This point should never be reached
        raise IndexError("OrderedSet index out of range")

    def add(self, value: T) -> None:
        if value not in self._dict:
            self._dict[value] = None

    def discard(self, value: T) -> None:
        self._dict.pop(value, None)

    def update(self, iterable: Iterable[T]) -> None:
        for x in iterable:
            self.add(x)

    def first(self) -> T | None:
        for x in self._dict:
            return x
        return None

    def try_replace(self, old: T, new: T) -> None:
        """Replace a potentially existing item with a new one, keeping its position if possible."""
        if new in self._dict or old not in self._dict:
            return
        self._dict = { (new if k == old else k): None for k in self._dict }

    def __repr__(self) -> str:
        items = ", ".join(repr(x) for x in self._dict)
        return f"OrderedSet([{items}])"

#------------------------------------------------------
# KeyedDict
#------------------------------------------------------

K = TypeVar("K")
V = TypeVar("V")

class KeyedDict(Generic[K, V]):
    __slots__ = ("_dict", "_key")
    _dict: dict[Hashable, tuple[K, V]]
    _key: Callable[[K], Hashable]

    def __init__(self, key: Callable[[K], Hashable], iterable: Iterable[tuple[K, V]] | None = None) -> None:
        self._dict = {}
        self._key = key
        for k, v in iterable or ():
            self._dict[self._key(k)] = (k, v)

    def __setitem__(self, k: K, v: V) -> None:
        self._dict[self._key(k)] = (k, v)

    def __getitem__(self, k: K) -> V:
        return self._dict[self._key(k)][1]

    def __delitem__(self, k: K) -> None:
        del self._dict[self._key(k)]

    def __contains__(self, k: object) -> bool:
        try: return self._key(k) in self._dict  # type: ignore[arg-type]
        except Exception: return False

    def __len__(self) -> int:
        return len(self._dict)

    def get(self, k: K, default: V | None = None) -> V | None:
        entry = self._dict.get(self._key(k))
        return default if entry is None else entry[1]

    def items(self) -> Iterator[tuple[K, V]]:
        yield from self._dict.values()

    def values(self) -> Iterator[V]:
        for _, v in self._dict.values():
            yield v

    def update(self, iterable: KeyedDict[K, V]|Iterable[tuple[K, V]]) -> None:
        if isinstance(iterable, KeyedDict):
            self._dict.update(iterable._dict)
        else:
            for k, v in iterable:
                self._dict[self._key(k)] = (k, v)


#------------------------------------------------------
# KeyedSet
#------------------------------------------------------

class KeyedSet(Generic[K], MutableSet[K]):
    __slots__ = ("_dict", "_key")
    _dict: dict[Hashable, K]
    _key: Callable[[K], Hashable]

    def __init__(self, key: Callable[[K], Hashable], iterable: Iterable[K] | None = None) -> None:
        self._dict = {}
        self._key = key
        if iterable is not None:
            self.update(iterable)

    def __contains__(self, item: K) -> bool:  # type: ignore[override]
        return self._key(item) in self._dict

    def __iter__(self) -> Iterator[K]:
        return iter(self._dict.values())

    def __len__(self) -> int:
        return len(self._dict)

    def add(self, value: K) -> None:
        if self._key(value) not in self._dict:
            self._dict[self._key(value)] = value

    def discard(self, value: K) -> None:
        self._dict.pop(self._key(value), None)

    def update(self, iterable: Iterable[K]) -> None:
        for x in iterable:
            self.add(x)

    def first(self) -> K | None:
        for _, v in self._dict.items():
            return v
        return None

    def __repr__(self) -> str:
        items = ", ".join(repr(x) for x in self._dict)
        return f"KeyedSet([{items}])"





import warnings

from v0.relationalai.semantics.metamodel.util import (NameCache, OrderedSet, FrozenOrderedSet, ordered_set, rewrite_list,
                                                   rewrite_set, filter_by_type, group_by, frozen, flatten_tuple, Printer,
                                                   split_by)

__all__ = ['NameCache', 'OrderedSet', 'FrozenOrderedSet', 'ordered_set', 'rewrite_list', 'rewrite_set', 'filter_by_type',
           'group_by', 'frozen', 'flatten_tuple', 'Printer', 'split_by']

warnings.warn(
    "relationalai.early_access.metamodel.util is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.util",
    DeprecationWarning,
    stacklevel=2,
)



# This script runs a QB program and captures high-level timing information about different parts of the system.
# The output is a JSON file containing timing information that can be used to analyze performance.

import argparse
import os
import json

from v0.relationalai.clients.resources.snowflake import Resources as snowflake_api, APP_NAME
from v0.relationalai.semantics.lqp.executor import LQPExecutor
from v0.relationalai.semantics.internal import internal
from v0.relationalai.clients.resources.snowflake.use_index_poller import UseIndexPoller as index_poller
from snowflake.connector.cursor import DictCursor

from enum import Enum

import time

# Categories that spans can belong to
class SpanCategory(Enum):
    DATA_PREP = "Data Preparation"
    QB_COMPILE = "QB Compilation"
    EXEC = "RAICode Execution"
    SNOWFLAKE = "Snowflake"
    RESULTS = "Result Processing"
    OTHER = "Other"
    UNKNOWN = "Unknown"

# Contextual information that describes an instrumented method. These are used to create spans.
#
# To better distinguish important spans, we have an `is_core` flag to indicate "core methods".
# Core methods represent high-level concepts that make sense to be reasoned about in isolation.
#
# Core methods must be mutually exclusive. This means that they should not overlap in time with one another,
# and should represent distinct concepts.
# For example, `exec_rai_app` and `process_results` are core methods since they are distinct high-level concepts
# that aid with understanding the overall time breakdown and do not overlap with one another.
# On the other hand, `exec_snowflake` is not a core method since it overlaps with other core methods and
# is tricky to reason about in isolation without knowing /what/ was being executed.
#
# Ideally, the sum of all core methods should equal the total time of the root span.
#
# Additionally, we have a `count_for_cat` flag to indicate whether this method should count towards the time of
# its category. This prevents duplicate counting for overlapping methods in the same category.
class TimeMarker:
    def __init__(self, name, category: SpanCategory, is_core=False, count_for_cat=False):
        self.name = name
        self.category = category
        self.is_core = is_core
        self.count_for_cat = count_for_cat

# Time span for a specific operation.
class TimeSpan:
    def __init__(self, id: str, marker: TimeMarker, parent_id, tags={}):
        self.id = id
        self.marker = marker
        self.parent_id = parent_id
        self.tags = tags

        self.start_time = None
        self.end_time = None

    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()

    def duration(self):
        assert self.start_time is not None, "Span has not started yet"
        assert self.end_time is not None, "Span has not ended yet"
        return self.end_time - self.start_time

# Context for the benchmarking. Keeps track of active spans and finished spans.
class BenchmarkContext:
    def __init__(self):
        self.current_spans = []
        self.finished_spans = []
        self.span_id_counter = 0
        self.print_logs = True
        self.gap_size = 0.0
        self.last_end_time = 0

        self.snowflake_calls = []
        self.last_qid = None
        self.qid_to_txnid = dict()

    def start_span(self, marker: TimeMarker, tags={}):
        parent_id = self.current_spans[-1].id if self.current_spans else None
        span_id = f"span_{self.span_id_counter}"
        self.span_id_counter += 1

        new_span = TimeSpan(span_id, marker, parent_id, tags=tags)
        new_span.start()
        self.current_spans.append(new_span)

        if new_span.marker.is_core:
            assert new_span.start_time is not None, "Span has not started yet"
            last_end_diff = new_span.start_time - self.last_end_time
            self.gap_size += last_end_diff
            if last_end_diff > 0.01:
                print(f"\tWarning: Gap of {last_end_diff:.4f} seconds before start of span `{new_span.id}` ({new_span.marker.name})")

        return new_span

    def end_span(self):
        assert self.current_spans, "No active span to end"
        span = self.current_spans.pop()
        span.end()
        self.finished_spans.append(span)
        if span.marker.is_core:
            self.last_end_time = span.end_time

        if self.print_logs:
            duration = span.duration()
            print(f"Span {span.id} ({span.marker.name}) ended, duration: {duration:.4f} seconds")
        return span

    def add_snowflake_call(self, qid: str, code: str):
        self.snowflake_calls.append({
            "sf_query_id": qid,
            "code": code
        })
        self.last_qid = qid
        if self.print_logs:
            print(f"Snowflake call [{qid}]: {code}")

    def add_txn_id(self, txn_id: str):
        assert self.last_qid is not None, "No Snowflake query ID recorded yet"
        assert self.last_qid not in self.qid_to_txnid, f"Snowflake query ID {self.last_qid} already has a transaction ID recorded"
        self.qid_to_txnid[self.last_qid] = txn_id
        self.last_qid = None

BENCHMARK_CTX = BenchmarkContext()

# Wrapper for instrumenting a method.
# If more fine-grained control is needed, consider monkey-patching manually with spans instead.
# E.g. if you want to add tags or have more complex logic.
class InstrumentedMethod:
    def __init__(self, module, method_name, category, is_core=False, count_for_cat=False, alias=None, override=None):
        self.module = module
        self.method_name = method_name

        marker_name = alias if alias else method_name
        self.marker = TimeMarker(marker_name, category, is_core, count_for_cat)

        self.original_method = getattr(module, method_name)
        if override:
            self.new_method = override(BENCHMARK_CTX, self.original_method, self.marker)
        else:
            self.new_method = self._create_instrumented_method()

    def _create_instrumented_method(self):
        def instrumented_method(*args, **kwargs):
            global BENCHMARK_CTX
            BENCHMARK_CTX.start_span(self.marker)
            result = self.original_method(*args, **kwargs)
            BENCHMARK_CTX.end_span()
            return result
        return instrumented_method

    def apply(self):
        setattr(self.module, self.method_name, self.new_method)

    def restore(self):
        setattr(self.module, self.method_name, self.original_method)

SF_COMPILATION_MARKER = TimeMarker("sf_compilation", SpanCategory.SNOWFLAKE)
SF_EXECUTION_MARKER = TimeMarker("sf_execution", SpanCategory.SNOWFLAKE)
SF_REQ_OVERHEAD_MARKER = TimeMarker("sf_request_overhead", SpanCategory.SNOWFLAKE)
def _exec_snowflake_override(bench_ctx, old_func, marker):
    def new_func(self, code, params, raw=False):
        cur = self._session.connection.cursor(DictCursor)
        try:
            cur.execute(code.replace(APP_NAME, self.get_app_name()), params)
            rows = cur.fetchall()
            qid = str(getattr(cur, "sfqid", None))
            assert qid is not None, "Snowflake query ID was not available"
        finally:
            cur.close()
        bench_ctx.add_snowflake_call(qid, code)

        # ===
        # Add dummy spans that will be populated later
        bench_ctx.start_span(SF_REQ_OVERHEAD_MARKER, tags={"sf_query_id": qid})
        bench_ctx.end_span()
        bench_ctx.start_span(SF_COMPILATION_MARKER, tags={"sf_query_id": qid})
        bench_ctx.end_span()
        bench_ctx.start_span(SF_EXECUTION_MARKER, tags={"sf_query_id": qid})
        bench_ctx.end_span()
        # ===

        return rows
    return new_func

def _is_dummy_sf_span(span):
    return span.marker in {SF_COMPILATION_MARKER, SF_EXECUTION_MARKER, SF_REQ_OVERHEAD_MARKER}

def _download_results_override(bench_ctx, old_func, marker):
    def new_func(self, artifact_info, txn_id: str, state):
        result = old_func(self, artifact_info, txn_id, state)
        bench_ctx.add_txn_id(txn_id)
        print("txn id: ", txn_id)
        return result
    return new_func

# List of all methods to track.
INSTRUMENTED_METHODS = [
    # Core methods. They must be mutually exclusive, high-level concepts, and ideally sum up to the total time.
    InstrumentedMethod(LQPExecutor, "prepare_data", SpanCategory.DATA_PREP, is_core=True, count_for_cat=True),
    InstrumentedMethod(LQPExecutor, "compile_lqp", SpanCategory.QB_COMPILE, is_core=True, count_for_cat=True),
    InstrumentedMethod(snowflake_api, "_exec_rai_app", SpanCategory.EXEC, is_core=True, count_for_cat=True, alias="exec_rai_app"),
    InstrumentedMethod(snowflake_api, "_fetch_exec_async_artifacts", SpanCategory.RESULTS, is_core=True, count_for_cat=True, alias="fetch_artifacts"),
    InstrumentedMethod(LQPExecutor, "_process_results", SpanCategory.RESULTS, is_core=True, count_for_cat=True, alias="process_results"),

    # Other trackers, these should also be on the high-level side but can overlap
    InstrumentedMethod(LQPExecutor, "execute", SpanCategory.EXEC, alias="execute_lqp"),
    InstrumentedMethod(snowflake_api, "_exec", SpanCategory.SNOWFLAKE, count_for_cat=True, alias="exec_snowflake"),
    InstrumentedMethod(index_poller, "_poll_loop", SpanCategory.DATA_PREP, alias="poll_once"),
    InstrumentedMethod(index_poller, "poll", SpanCategory.DATA_PREP, alias="poll_main"),
    InstrumentedMethod(snowflake_api, "_exec_sql", SpanCategory.SNOWFLAKE, alias="exec_sql", override=_exec_snowflake_override),
    InstrumentedMethod(snowflake_api, "_build_snowflake_session", SpanCategory.SNOWFLAKE),
    InstrumentedMethod(snowflake_api, "_download_results", SpanCategory.RESULTS, override=_download_results_override),
]


def _make_unknown_span(parent_id, parent_duration, known_children_duration):
    unknown_duration = parent_duration - known_children_duration
    if unknown_duration <= 0.01:
        return None
    return {
        "name": "unknown",
        "duration": unknown_duration,
        "category": SpanCategory.UNKNOWN.value,
        "parent": parent_id,
        "tags": {},
        "is_core": True,
        "children": [],
        "total_percent": (unknown_duration / parent_duration) * 100,
        "parent_percent": (unknown_duration / parent_duration) * 100,
    }

def _make_task_times_dict(spans, root_span):
    total_duration = root_span.duration()
    task_times = dict()
    for span in spans:
        if span.marker.name not in task_times:
            task_times[span.marker.name] = {"duration": 0.0, "category": span.marker.category.value}
        task_times[span.marker.name]["duration"] += span.duration()

    # Add total percent to each
    for v in task_times.values():
        v["percent"] = (v["duration"] / total_duration) * 100
    return task_times

def _make_spans_dict(spans, root_span):
    spans_dict = dict()
    for span in spans:
        spans_dict[span.id] = {
            "name": span.marker.name,
            "duration": span.duration(),
            "category": span.marker.category.value,
            "parent": span.parent_id,
            "tags": span.tags,
            "is_core": span.marker.is_core,

            # To be filled later
            "children": [],
        }

    # Fill in children
    for span in spans:
        if span.parent_id:
            spans_dict[span.parent_id]["children"].append(span.id)

    # Add total_percent and parent_percent
    total_duration = root_span.duration()
    for span in spans:
        span_info = spans_dict[span.id]
        span_info["total_percent"] = (span_info["duration"] / total_duration) * 100
        if span.parent_id:
            parent_duration = spans_dict[span.parent_id]["duration"]
            span_info["parent_percent"] = (span_info["duration"] / parent_duration) * 100
        else:
            # Root span has no parent
            span_info["parent_percent"] = 100.0

    # Add "Unknown" spans for when sum of children < parent
    for span in spans:
        children = spans_dict[span.id]["children"]
        if not children:
            continue
        known_children_duration = sum(spans_dict[child_id]["duration"] for child_id in children)
        unknown_span = _make_unknown_span(span.id, spans_dict[span.id]["duration"], known_children_duration)
        if unknown_span:
            unknown_id = f"{span.id}_unknown"
            spans_dict[unknown_id] = unknown_span
            spans_dict[span.id]["children"].append(unknown_id)

    return spans_dict

def _make_core_times_dict(spans, root_span):
    core_times = dict()
    total_duration = root_span.duration()
    duration_accounted = 0.0
    for span in spans:
        if span.marker.is_core:
            if span.marker.name not in core_times:
                core_times[span.marker.name] = {
                    "duration": 0.0,
                }
            core_times[span.marker.name]["duration"] += span.duration()
            duration_accounted += span.duration()
    for v in core_times.values():
        v["total_percent"] = (v["duration"] / total_duration) * 100

    if duration_accounted < total_duration - 0.01:
        core_times["unknown"] = {
            "duration": total_duration - duration_accounted,
            "total_percent": ((total_duration - duration_accounted) / total_duration) * 100
        }
    return core_times

def _make_category_times_dict(spans, root_span):
    category_times = dict()
    for span in spans:
        if span.marker.count_for_cat:
            cat_name = span.marker.category.value
            if cat_name not in category_times:
                category_times[cat_name] = {
                    "duration": 0.0,
                }
            category_times[cat_name]["duration"] += span.duration()
    return category_times

def main():
    parser = argparse.ArgumentParser(description="Extract LQP requests to run locally")
    parser.add_argument("--program", type=str, help="Path to the QB program to run", required=True)
    parser.add_argument("--outdir", type=str, help="Path to the output directory", required=True)

    args = parser.parse_args()
    program = args.program
    outdir = args.outdir

    if not os.path.isfile(program):
        raise ValueError(f"Program path {program} is not a file")
    if not os.path.isdir(outdir):
        raise ValueError(f"Output path {outdir} is not a directory")

    benchmark_lqp(program, outdir)

def _run_instrumented_qb(qb_prog_path: str):
    global BENCHMARK_CTX
    global INSTRUMENTED_METHODS

    # Run the QB program with instrumentation
    for method in INSTRUMENTED_METHODS:
        method.apply()

    # Special marker for the root span
    e2e_marker = TimeMarker("e2e", SpanCategory.OTHER)
    root_span = BENCHMARK_CTX.start_span(e2e_marker)
    BENCHMARK_CTX.last_end_time = root_span.start_time
    with internal.with_overrides(**{
        'dry_run': False,
        'reasoner.rule.use_lqp': True,
    }):
        with open(qb_prog_path) as f:
            # Compile with the correct filename
            code = compile(f.read(), qb_prog_path, "exec")
            exec(code)
    root_span = BENCHMARK_CTX.end_span()
    assert root_span.marker == e2e_marker, "Mismatched root span"
    for method in INSTRUMENTED_METHODS:
        method.restore()
    spans = BENCHMARK_CTX.finished_spans
    return spans, root_span

def get_sf_query_info(bench_ctx):
    snowflake_calls = bench_ctx.snowflake_calls
    qids = []
    for call in snowflake_calls:
        qids.append(call["sf_query_id"])
    assert len(qids) == len(snowflake_calls), "Duplicate Snowflake query IDs found"

    result = dict()

    qids_to_info = _get_query_info(qids)
    for call in snowflake_calls:
        qid = call["sf_query_id"]
        assert qid in qids_to_info, f"Missing info for Snowflake query ID {qid}"
        qid_info = qids_to_info[qid]
        call["sf_compilation_time"] = qid_info["compilation_time"]
        call["sf_execution_time"] = qid_info["execution_time"]
        if qid in bench_ctx.qid_to_txnid:
            call["rai_txn_id"] = bench_ctx.qid_to_txnid[qid]
        result[qid] = call
    return result

def _get_query_info(qids):
    from v0.relationalai.clients.resources.snowflake import Resources as snowflake_client
    client = snowflake_client()

    qids_str = "','".join(qids)
    timing_query = f"""
        SELECT
            query_id,
            start_time,
            end_time,
            total_elapsed_time,
            compilation_time,
            execution_time,
            queued_overload_time,
            queued_provisioning_time,
            queued_repair_time
        FROM table(snowflake.information_schema.query_history())
        WHERE query_id IN ('{qids_str}')
        ORDER BY start_time;
        """
    results = client._exec(timing_query)
    if len(results) != len(qids):
        print(f"Expected {len(qids)} results, but got {len(results)}")
        print("Trying again after a few more seconds...")
        time.sleep(5)
        results = client._exec(timing_query)
    assert len(results) == len(qids), f"Expected {len(qids)} results, but got {len(results)}"

    qid_to_info = dict()
    for row in results:
        qid = row["QUERY_ID"]
        qid_to_info[qid] = {
            "compilation_time": row["COMPILATION_TIME"]/1000.0,
            "execution_time": row["EXECUTION_TIME"]/1000.0,

        }
    return qid_to_info

def populate_dummy_spans(spans, sf_query_info):
    parentid_to_qid = dict()
    for span in spans:
        if not _is_dummy_sf_span(span):
            continue
        parentid_to_qid[span.parent_id] = span.tags["sf_query_id"]
    assert len(parentid_to_qid) == len(sf_query_info), "Mismatch between dummy spans and Snowflake query info"

    parentid_to_times = dict()
    for span in spans:
        if span.id in parentid_to_qid:
            parentid_to_times[span.id] = span.duration()

    for span in spans:
        if not _is_dummy_sf_span(span):
            continue

        qid = span.tags.get("sf_query_id", None)
        assert qid is not None, f"Span {span.id} ({span.marker.name}) is missing sf_query_id tag"

        info = sf_query_info.get(qid, None)
        assert info is not None, f"Missing Snowflake info for query ID {qid}"

        comp_time = info["sf_compilation_time"]
        exec_time = info["sf_execution_time"]

        total_parent_time = parentid_to_times[span.parent_id]
        # We order it like: compilation -> execution -> overhead
        # The overhead is the remaining time after accounting for the snowflake-reported times
        if span.marker.name == "sf_compilation":
            span.start_time = span.start_time
            span.end_time = span.start_time + comp_time
        elif span.marker.name == "sf_execution":
            offset = span.start_time + comp_time
            span.start_time = offset
            span.end_time = offset + exec_time
        elif span.marker.name == "sf_request_overhead":
            offset = span.start_time + comp_time + exec_time
            duration = total_parent_time - (comp_time + exec_time)
            span.start_time = offset
            span.end_time = offset + duration
        else:
            raise ValueError(f"Unknown marker name {span.marker.name}")
        span.tags["rai_txn_id"] = info.get("rai_txn_id", None)
        assert span.end_time >= span.start_time, f"Span {span.id} ({span.marker.name}) has negative duration after population"

def _make_snowflake_calls_dict(sf_query_info):
    result = []
    for qid, info in sf_query_info.items():
        result.append(info)
    return result

def _make_metadata_dict(bench_ctx, root_span, prog_path):
    metadata = []
    metadata.append({
        "key": "total_time",
        "alias": "Total Time",
        "value": root_span.duration(),
    })
    metadata.append({
        "key": "gap_time",
        "alias": "Untracked Time",
        "value": bench_ctx.gap_size,
    })
    metadata.append({
        "key": "program",
        "alias": "QB Program",
        "value": prog_path,
    })
    return metadata

def benchmark_lqp(qb_prog_path: str, outdir: str):
    global BENCHMARK_CTX
    spans, root_span = _run_instrumented_qb(qb_prog_path)

    print(f"Total benchmark time: {root_span.duration():.4f} seconds")
    print(f"\tMissing time not tracked by core methods: {BENCHMARK_CTX.gap_size:.4f} seconds")

    sf_query_info = get_sf_query_info(BENCHMARK_CTX)
    populate_dummy_spans(spans, sf_query_info)

    # Write results to file
    final_json_dict = dict()
    final_json_dict["spans"] = _make_spans_dict(spans, root_span)
    final_json_dict["task_times"] = _make_task_times_dict(spans, root_span)
    final_json_dict["core_times"] = _make_core_times_dict(spans, root_span)
    final_json_dict["category_times"] = _make_category_times_dict(spans, root_span)
    final_json_dict["snowflake_calls"] = _make_snowflake_calls_dict(sf_query_info)
    final_json_dict["metadata"] = _make_metadata_dict(BENCHMARK_CTX, root_span, qb_prog_path)

    final_json = json.dumps(final_json_dict, default=str)
    out_path = os.path.join(outdir, "benchmark_results.json")
    with open(out_path, "w") as f:
        f.write(final_json)

    print(f"Wrote benchmark results to {out_path}")

if __name__ == "__main__":
    main()



from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from v0.relationalai.semantics.metamodel.typer import Checker, InferTypes, typer
from v0.relationalai.semantics.metamodel import helpers, types
from v0.relationalai.semantics.metamodel.util import FrozenOrderedSet

from v0.relationalai.semantics.metamodel.rewrite import Flatten

from ..metamodel.rewrite import DNFUnionSplitter, ExtractNestedLogicals, FormatOutputs
from .rewrite import (
    AnnotateConstraints, CDC, ExtractCommon, ExtractKeys, FunctionAnnotations, QuantifyVars,
    Splinter, SplitMultiCheckRequires
)
from v0.relationalai.semantics.lqp.utils import output_names

from typing import cast, List, Sequence, Tuple, Union, Optional, Iterable
from collections import defaultdict
import pandas as pd
import hashlib

def lqp_passes() -> list[Pass]:
    return [
        SplitMultiCheckRequires(),
        FunctionAnnotations(),
        AnnotateConstraints(),
        Checker(),
        CDC(), # specialize to physical relations before extracting nested and typing
        ExtractNestedLogicals(), # before InferTypes to avoid extracting casts
        InferTypes(),
        DNFUnionSplitter(),
        ExtractKeys(),
        FormatOutputs(),
        ExtractCommon(), # Extracts tasks that will become common after Flatten into their own definition
        Flatten(),
        Splinter(), # Splits multi-headed rules into multiple rules
        QuantifyVars(), # Adds missing existentials
        EliminateData(),  # Turns Data nodes into ordinary relations.
        DeduplicateVars(),  # Deduplicates vars in Updates and Outputs.
        PeriodMath(),  # Rewrite date period uses.
        ConstantsToVars(),  # Turns constants in Updates and Outputs into vars.
        UnifyDefinitions(),
    ]

# LQP does not support multiple definitions for the same relation. This pass unifies all
# definitions for each relation into a single definition using a union.
class UnifyDefinitions(Pass):
    def __init__(self):
        super().__init__()

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # Maintain a cache of renamings for each relation. These need to be consistent
        # across all definitions of the same relation.
        self.renamed_relation_args: dict[Union[ir.Value, ir.Relation], list[ir.Var]] = {}

        root = cast(ir.Logical, model.root)
        new_tasks = self.get_combined_multidefs(root)
        return ir.Model(
            model.engines,
            model.relations,
            model.types,
            f.logical(
                tuple(new_tasks),
                root.hoisted,
                root.engine,
            ),
            model.annotations,
        )

    def _get_heads(self, logical: ir.Logical) -> list[Union[ir.Update, ir.Output]]:
        derives = []
        for task in logical.body:
            if isinstance(task, ir.Update) and task.effect == ir.Effect.derive:
                derives.append(task)
            elif isinstance(task, ir.Output):
                derives.append(task)
        return derives

    def _get_non_heads(self, logical: ir.Logical) -> list[ir.Task]:
        non_derives = []
        for task in logical.body:
            if not(isinstance(task, ir.Update) and task.effect == ir.Effect.derive) and not isinstance(task, ir.Output):
                non_derives.append(task)
        return non_derives

    def _get_head_identifier(self, head: Union[ir.Update, ir.Output]) -> Optional[ir.Value]:
        if isinstance(head, ir.Update):
            return head.relation
        else:
            assert isinstance(head, ir.Output)
            if len(head.aliases) <= 2:
                # For processing here, we need output to have at least the column markers
                # `cols` and `col`, and also a key
                return None

            output_alias_names = helpers.output_alias_names(head.aliases)
            output_vals = helpers.output_values(head.aliases)

            # For normal outputs, the pattern is output[keys](cols, "col000" as 'col', ...)
            if output_alias_names[0] == "cols" and output_alias_names[1] == "col":
                return output_vals[1]

            # For exports, the pattern is output[keys]("col000" as 'col', ...)
            if rel_builtins.export_annotation in head.annotations:
                if output_alias_names[0] == "col":
                    return output_vals[0]

        return None

    def get_combined_multidefs(self, root: ir.Logical) -> list[ir.Logical]:
        # Step 1: Group tasks by the relation they define.
        relation_to_tasks: dict[Union[None, ir.Value, ir.Relation], list[ir.Logical]] = defaultdict(list)

        for task in root.body:
            task = cast(ir.Logical, task)
            task_heads = self._get_heads(task)

            # Some relations do not need to be grouped, e.g., if they don't contain a
            # derive. Use `None` as a placeholder key for these cases.
            if len(task_heads) != 1:
                relation_to_tasks[None].append(task)
                continue

            head_id = self._get_head_identifier(task_heads[0])
            relation_to_tasks[head_id].append(task)

        # Step 2: For each relation, combine all of the body definitions into a union.
        result_tasks = []
        for relation, tasks in relation_to_tasks.items():
            # If there's only one task for the relation, or if grouping is not needed, then
            # just keep the original tasks.
            if len(tasks) == 1 or relation is None:
                result_tasks.extend(tasks)
                continue

            result_tasks.append(self._combine_tasks_into_union(tasks))
        return result_tasks

    def _get_variable_mapping(self, logical: ir.Logical) -> dict[ir.Value, ir.Var]:
        heads = self._get_heads(logical)
        assert len(heads) == 1, "should only have one head in a logical at this stage"
        head = heads[0]

        var_mapping = {}
        head_id = self._get_head_identifier(head)

        if isinstance(head, ir.Update):
            args_for_renaming = head.args
        else:
            assert isinstance(head, ir.Output)
            output_alias_names = helpers.output_alias_names(head.aliases)
            if output_alias_names[0] == "cols" and output_alias_names[1] == "col":
                assert len(head.aliases) > 2

                # For outputs, we do not need to rename the `cols` and `col` markers or the
                # keys.
                output_values = helpers.output_values(head.aliases)[2:]

            else:
                assert rel_builtins.export_annotation in head.annotations and output_alias_names[0] == "col"
                assert len(head.aliases) > 1

                # For exports, we do not need to rename the `col` marker or the keys.
                output_values = helpers.output_values(head.aliases)[1:]

            args_for_renaming = []
            for v in output_values:
                if head.keys and isinstance(v, ir.Var) and v in head.keys:
                    continue
                args_for_renaming.append(v)

        if head_id not in self.renamed_relation_args:
            renamed_vars = []
            for (i, arg) in enumerate(args_for_renaming):
                typ = typer.to_type(arg)
                assert arg not in var_mapping, "args of update should be unique"
                if isinstance(arg, ir.Var):
                    var_mapping[arg] = ir.Var(typ, arg.name)
                else:
                    var_mapping[arg] = ir.Var(typ, f"arg_{i}")

                renamed_vars.append(var_mapping[arg])
            self.renamed_relation_args[head_id] = renamed_vars
        else:
            for (arg, var) in zip(args_for_renaming, self.renamed_relation_args[head_id]):
                var_mapping[arg] = var

        return var_mapping

    def _rename_variables(self, logical: ir.Logical) -> ir.Logical:
        class RenameVisitor(visitor.Rewriter):
            def __init__(self, var_mapping: dict[ir.Value, ir.Var]):
                super().__init__()
                self.var_mapping = var_mapping

            def _get_mapped_value(self, val: ir.Value) -> ir.Value:
                if isinstance(val, tuple):
                    return tuple(self._get_mapped_value(t) for t in val)
                return self.var_mapping.get(val, val)

            def _get_mapped_values(self, vals: Iterable[ir.Value]) -> list[ir.Value]:
                return [self._get_mapped_value(v) for v in vals]

            def handle_var(self, node: ir.Var, parent: ir.Node) -> ir.Var:
                return self.var_mapping.get(node, node)

            # TODO: ideally, extend the rewriter class to allow rewriting PyValue to Var so
            # we don't need to separately handle all cases containing them.
            def handle_update(self, node: ir.Update, parent: ir.Node) -> ir.Update:
                return ir.Update(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.effect,
                    node.annotations,
                )

            def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
                return ir.Lookup(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.annotations,
                )

            def handle_output(self, node: ir.Output, parent: ir.Node) -> ir.Output:
                new_aliases = FrozenOrderedSet(
                    [(name, self._get_mapped_value(value)) for name, value in node.aliases]
                )
                if node.keys:
                    new_keys = FrozenOrderedSet(
                        [self.var_mapping.get(key, key) for key in node.keys]
                    )
                else:
                    new_keys = node.keys

                return ir.Output(
                    node.engine,
                    new_aliases,
                    new_keys,
                    node.annotations,
                )

            def handle_construct(self, node: ir.Construct, parent: ir.Node) -> ir.Construct:
                new_values = tuple(self._get_mapped_values(node.values))
                new_id_var = self.var_mapping.get(node.id_var, node.id_var)
                return ir.Construct(
                    node.engine,
                    new_values,
                    new_id_var,
                    node.annotations,
                )

            def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node) -> ir.Aggregate:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self._get_mapped_values(node.args))
                return ir.Aggregate(
                    node.engine,
                    node.aggregation,
                    new_projection,
                    new_group,
                    new_args,
                    node.annotations,
                )

            def handle_rank(self, node: ir.Rank, parent: ir.Node) -> ir.Rank:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self.var_mapping.get(arg, arg) for arg in node.args)
                new_result = self.var_mapping.get(node.result, node.result)

                return ir.Rank(
                    node.engine,
                    new_projection,
                    new_group,
                    new_args,
                    node.arg_is_ascending,
                    new_result,
                    node.limit,
                    node.annotations,
                )

        var_mapping = self._get_variable_mapping(logical)

        renamer = RenameVisitor(var_mapping)
        result = renamer.walk(logical)

        # Also need to append the equality for each renamed constant. E.g., if the mapping
        # contains (50.0::FLOAT -> arg_2::FLOAT), we need to add
        # `eq(arg_2::FLOAT, 50.0::FLOAT)` to the result.
        value_eqs = []
        for (old_var, new_var) in var_mapping.items():
            if not isinstance(old_var, ir.Var):
                value_eqs.append(f.lookup(rel_builtins.eq, [new_var, old_var]))

        return ir.Logical(
            result.engine,
            result.hoisted,
            tuple(value_eqs) + tuple(result.body),
            result.annotations,
        )

    # This function is the main workhorse for this rewrite pass. It takes a list of tasks
    # that define the same relation, and combines them into a single task that defines
    # the relation using a union of all of the bodies.
    def _combine_tasks_into_union(self, tasks: list[ir.Logical]) -> ir.Logical:
        # Step 1: Rename the variables in all tasks so that they will match the final derive
        # after reconstructing into a union
        renamed_tasks = [self._rename_variables(task) for task in tasks]

        # Step 2: Get the final derive
        derives = self._get_heads(renamed_tasks[0])
        assert len(derives) == 1, "should only have one derive in a logical at this stage"
        # Also make sure that all the derives are the same. This should be the case because
        # we renamed all the variables to be the same in step 1.
        for task in renamed_tasks[1:]:
            assert self._get_heads(task) == derives, "all derives should be the same"

        derive = derives[0]

        # Step 3: Remove the final `derive` from each task
        renamed_task_bodies = [
            f.logical(
                tuple(self._get_non_heads(t)),  # Only keep non-head tasks
                t.hoisted,
                t.engine,
            )
            for t in renamed_tasks
        ]

        # Step 4: Construct a union of all the task bodies
        union = f.union(
            tuple(renamed_task_bodies),
            [],
            renamed_tasks[0].engine,
        )

        # Step 5: Add the final derive back
        return f.logical(
            (union, derive),
            [],
            renamed_tasks[0].engine,
        )

# Creates intermediary relations for all Data nodes and replaces said Data nodes
# with a Lookup into these created relations. Reuse duplicate created relations.
class EliminateData(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.DataRewriter()
        return r.walk(model)

    # Does the actual work.
    class DataRewriter(visitor.Rewriter):
        new_relations: list[ir.Relation]
        new_updates: list[ir.Logical]
        # Counter for naming new relations.
        # It must be that new_count == len new_updates == len new_relations.
        new_count: int
        # Cache for Data nodes to avoid creating duplicate intermediary relations
        data_cache: dict[str, ir.Relation]

        def __init__(self):
            self.new_relations = []
            self.new_updates = []
            self.new_count = 0
            self.data_cache = {}
            super().__init__()

        # Create a cache key for a Data node based on its structure and content
        def _data_cache_key(self, node: ir.Data) -> str:
            values = pd.util.hash_pandas_object(node.data).values
            return hashlib.sha256(bytes(values)).hexdigest()

        def _intermediary_relation(self, node: ir.Data) -> ir.Relation:
            cache_key = self._data_cache_key(node)
            if cache_key in self.data_cache:
                return self.data_cache[cache_key]
            self.new_count += 1
            intermediary_name = f"formerly_Data_{self.new_count}"

            intermediary_relation = f.relation(
                intermediary_name,
                [f.field(v.name, v.type) for v in node.vars]
            )
            self.new_relations.append(intermediary_relation)

            intermediary_update = f.logical([
                # For each row (union), equate values and their variable (logical).
                f.union(
                    [
                        f.logical(
                            [
                                f.lookup(rel_builtins.eq, [f.literal(val, var.type), var])
                                for (val, var) in zip(row, node.vars)
                            ],
                        )
                        for row in node
                    ],
                    hoisted = node.vars,
                ),
                # And pop it back into the relation.
                f.update(intermediary_relation, node.vars, ir.Effect.derive),
            ])
            self.new_updates.append(intermediary_update)

            # Cache the result for reuse
            self.data_cache[cache_key] = intermediary_relation

            return intermediary_relation

        # Create a new intermediary relation representing the Data (and pop it in
        # new_updates/new_relations) and replace this Data with a Lookup of said
        # intermediary.
        def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Lookup:
            intermediary_relation = self._intermediary_relation(node)
            replacement_lookup = f.lookup(intermediary_relation, node.vars)

            return replacement_lookup

        # Walks the model for the handle_data work then updates the model with
        # the new state.
        def handle_model(self, model: ir.Model, parent: None):
            walked_model = super().handle_model(model, parent)
            assert len(self.new_relations) == len(self.new_updates) and self.new_count == len(self.new_relations)

            # This is okay because its LQP.
            assert isinstance(walked_model.root, ir.Logical)
            root_logical = cast(ir.Logical, walked_model.root)

            # We may need to add the new intermediaries from handle_data to the model.
            if self.new_count  == 0:
                return model
            else:
                return ir.Model(
                    walked_model.engines,
                    walked_model.relations | self.new_relations,
                    walked_model.types,
                    ir.Logical(
                        root_logical.engine,
                        root_logical.hoisted,
                        root_logical.body + tuple(self.new_updates),
                        root_logical.annotations,
                    ),
                    walked_model.annotations,
                )

# Deduplicate Vars in Updates and Outputs.
class DeduplicateVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.VarDeduplicator()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def dedup_values(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        # If a var is seen more than once, it is a duplicate and we will create
        # a new Var and equate it with the seen one.
        seen_vars = set()

        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            # Duplicates can only occur within Vars.
            # TODO: we don't know for sure if these are the only relevant cases.
            if isinstance(val, ir.Default) or isinstance(val, ir.Var):
                var = val if isinstance(val, ir.Var) else val.var
                if var in seen_vars:
                    new_var = ir.Var(var.type, var.name + "_dup_" + str(i))
                    new_val = new_var if isinstance(val, ir.Var) else ir.Default(new_var, val.value)
                    new_vals.append(new_val)
                    eqs.append(f.lookup(rel_builtins.eq, [new_var, var]))
                else:
                    seen_vars.add(var)
                    new_vals.append(val)
            else:
                # No possibility of problematic duplication.
                new_vals.append(val)

        return new_vals, eqs

    # Returns a reconstructed output with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_output(output: ir.Output) -> List[Union[ir.Output, ir.Lookup]]:
        vals = helpers.output_values(output.aliases)
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(vals)
        # Need the names so we can recombine.
        alias_names = output_names(output.aliases)
        new_output = ir.Output(
            output.engine,
            FrozenOrderedSet(list(zip(alias_names, deduped_vals))),
            output.keys,
            output.annotations,
        )
        return req_lookups + [new_output]

    # Returns a replacement update with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class VarDeduplicator(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update/handle_output
        # because in addition to modifying said update/output we require new
        # lookups (equality between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Output):
                    new_body.extend(DeduplicateVars.dedup_output(subtask))
                elif isinstance(subtask, ir.Update):
                    new_body.extend(DeduplicateVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )

# Generate date arithmetic expressions, such as
# `rel_primitive_date_add(:day, [date] delta, res_2)` by finding the period
# expression for the delta and adding the period type to the date arithmetic expression.
#
# date_add and it's kin are generated by a period expression, e.g.,
# `day(delta, res_1)`
# followed by the date arithmetic expression using the period
# `date_add([date] res_1 res_2)`
class PeriodMath(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        period_rewriter = self.PeriodRewriter()
        model = period_rewriter.walk(model)
        period_math_rewriter = self.PeriodMathRewriter(period_rewriter.period_vars)
        model = period_math_rewriter.walk(model)
        return model

    # Find all period builtins. We need to make them safe for the emitter (either by
    # translating to a cast, or removing) and store the variable and period type for use
    # in the date/datetime add/subtract expressions.
    class PeriodRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = {}

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name not in {
                "year", "month", "week", "day", "hour", "minute", "second", "millisecond", "microsecond", "nanosecond"
            }:
                return node

            assert len(node.args) == 2, "Expect 2 arguments for period builtins"
            assert isinstance(node.args[1], ir.Var), "Expect result to be a variable"
            period = node.relation.name
            result_var = node.args[1]
            self.period_vars[result_var] = period

            # Ideally we could now remove the unused and unhandled period type construction
            # but we may also need to cast the original variable to an Int64 for use by the
            # date/datetime add/subtract expressions.
            # TODO: Remove the node entirely where possible and update uses of the result
            return f.lookup(rel_builtins.cast, [types.Int64, node.args[0], result_var])

    # Update date/datetime add/subtract expressions with period information.
    class PeriodMathRewriter(visitor.Rewriter):
        def __init__(self, period_vars: dict[ir.Var, str]):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = period_vars

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name not in {
                "date_add", "date_subtract", "datetime_add", "datetime_subtract"
            }:
                return node

            if len(node.args) == 4:
                # We've already visited this lookup
                return node

            assert isinstance(node.args[1], ir.Var), "Expect period to be a variable"
            period_var = node.args[1]
            assert period_var in self.period_vars, "datemath found, but no vars to insert"

            period = self.period_vars[period_var]

            new_args = [f.literal(period, types.Symbol)] + [arg for arg in node.args]

            return f.lookup(node.relation, new_args)

# Rewrite constants to vars in Updates. This results in a more normalized format where
# updates contain only variables. This allows for easier rewrites in later passes.
class ConstantsToVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.ConstantToVarRewriter()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def replace_constants_with_vars(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            if isinstance(val, ir.PyValue) or isinstance(val, ir.Literal):
                # Replace constant with a new Var.
                typ = typer.to_type(val)
                assert isinstance(typ, ir.ScalarType), "can only replace scalar constants with vars"
                new_var = ir.Var(typ, f"{typ.name.lower()}_{i}")
                new_vals.append(new_var)
                eqs.append(f.lookup(rel_builtins.eq, [new_var, val]))
            else:
                new_vals.append(val)

        return new_vals, eqs

    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = ConstantsToVars.replace_constants_with_vars(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class ConstantToVarRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update because in
        # addition to modifying said update we require new lookups (equality
        # between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Update):
                    new_body.extend(ConstantsToVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )
